# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------
# Copyright (c) 2020
#
# See the LICENSE file for details
# see the AUTHORS file for authors
# ----------------------------------------------------------------------

#--------------------
# System wide imports
# -------------------

import os
import os.path
import logging
import datetime
import hashlib
import math
import re
import subprocess

try:
    # Python 2
    import ConfigParser
except:
    import configparser as ConfigParser


# ---------------------
# Third party libraries
# ---------------------

import exifread
import rawpy
import numpy as np
import jdcal

#--------------
# local imports
# -------------

from .           import DEF_CAMERA_TPL, DEF_TSTAMP
from .utils      import chop, Point, ROI
from .exceptions import ConfigError, MetadataError, TimestampError

# ----------------
# Module constants
# ----------------

# Array indexes in Bayer pattern
R1 = 0
G2 = 1
G3 = 2
B4 = 3


BG_X1 = 400
BG_Y1 = 200
BG_X2 = 550
BG_Y2 = 350

FRACTION_REGEXP = re.compile(r'(\d+)/(\d+)')


# -----------------------
# Module global variables
# -----------------------

log = logging.getLogger("azotea")

# ----------
# Exceptions
# ----------


# =======
# CLASSES
# =======

class CameraCache(object):

    def __init__(self, camerapath):
        self._points_cache = {}
        self._steps_cache = {}
        self._camerapath = camerapath


    def lookup(self, model):
        '''
        Load camera configuration from configuration file
        '''
        if model in self._points_cache.keys():
            return self._points_cache[model], self._steps_cache[model]

        if not (os.path.exists(self._camerapath)):
            self._camerapath = DEF_CAMERA_TPL

        parser  =  ConfigParser.RawConfigParser()
        # str is for case sensitive options
        parser.optionxform = str
        parser.read(self._camerapath)
        if not parser.has_section(model):
            raise ConfigError(model)

        r1 = chop(parser.get(model,"R1"),',')
        g2 = chop(parser.get(model,"G2"),',')
        g3 = chop(parser.get(model,"G3"),',')
        b4 = chop(parser.get(model,"B4"),',')

        points = [ 
            Point(x=int(r1[0]), y=int(r1[1])),
            Point(x=int(g2[0]), y=int(g2[1])),
            Point(x=int(g3[0]), y=int(g3[1])),
            Point(x=int(b4[0]), y=int(b4[1])),
        ]
        steps = [ int(r1[2]), int(g2[2]),  int(g3[2]), int(b4[2])]

        self._points_cache[model] = points
        self._steps_cache[model] = steps
        return points, steps


        


class CameraImage(object):

    ExiftoolFixed = False

    HEADERS = [
            'tstamp'         ,
            'name'           ,
            'model'          ,
            'iso'            ,
            'roi'            ,
            'dark_roi'       ,
            'exptime'        ,
            'focal_length'   ,
            'f_number'       ,
            'bias'           ,
            'type'           ,
            'aver_signal_R1',
            'vari_signal_R1',
            'aver_signal_G2',
            'vari_signal_G2',
            'aver_signal_G3',
            'vari_signal_G3',
            'aver_signal_B4',
            'vari_signal_B4',
        ]
                
               

    # ========== #
    # Public API #
    # ========== #

    def __init__(self, filepath, cache):
        self.filepath   = filepath
        self.cache      = cache # Camera reading parameters cache
        self.roi        = None  # foreground rectangular region where signal is estimated
        self.dkroi      = None  # dark rectangular region where bias is estimated
        self.exif       = None
        self.metadata   = {}  # Subset of EXIF metadata we are interested in
        self.name       = os.path.basename(self.filepath)
        self.signal     = []    # Bayer array for signal ROI
        self.dark       = []    # Bayer array for dark ROI
        self.path       = filepath  
        self.k          = [ Point(), Point(), Point(), Point()] # Array of Points to properly read each channel
        self.step       = [2, 2, 2, 2]
    

    def loadEXIF(self):
        '''Load EXIF metadata'''   
        #log.debug("%s: Loading EXIF metadata",self.name)
        with open(self.filepath, "rb") as f:
            logging.disable(logging.INFO)
            self.exif = exifread.process_file(f, details=False)
            logging.disable(logging.NOTSET)
        if not self.exif:
            self._exiftool()
            #raise MetadataError(self.filepath)

        self.metadata['name']         = self.name
        self.model                    = str(self.exif.get('Image Model'))
        self.metadata['model']        = self.model
        self.metadata['tstamp']       = self._iso8601(str(self.exif.get('Image DateTime', None)))
        self.metadata['iso']          = str(self.exif.get('EXIF ISOSpeedRatings', None))
        self.metadata['focal_length'] = self.getFocalLength()
        self.metadata['f_number']     = self.getFNumber()
        self.metadata['exptime'], self.metadata['type'] = self.getExposureTime()
        self.metadata['bias']         = 0   # Maybe one day we could extract is from EXIF
        self.metadata['night']        = self.night() 
        return self.metadata


    def getExposureTime(self):
        imagetyp = None
        exptime = str(self.exif.get('EXIF ExposureTime', None))
        exptime = self._fraction_to_float(exptime)
        if exptime < 1.0:
            log.warn("Image %s (t=%f) could serve as a BIAS image", self.name, exptime)
            imagetyp = "BIAS"
        return exptime, imagetyp


    def getFNumber(self):
        f_number = str(self.exif.get('EXIF FNumber', None))
        return self._fraction_to_float(f_number)


    def getFocalLength(self):
        temp = str(self.exif.get('EXIF FocalLength', None))
        if temp == '0':
            temp = None
        return self._fraction_to_float(temp)


    def read(self):
        '''Read RAW pixels''' 
        self._lookup()
        log.debug("%s: Loading RAW data from %s", self.name, self.model)
        img = rawpy.imread(self.filepath)
        log.debug("%s: Color description is %s", self.name, img.color_desc)
        # R1 channel
        self.signal.append(img.raw_image[self.k[R1].x::self.step[R1], self.k[R1].y::self.step[R1]])
        # G2 channel
        self.signal.append(img.raw_image[self.k[G2].x::self.step[G2], self.k[G2].y::self.step[G2]])
        # G3 channel
        self.signal.append(img.raw_image[self.k[G3].x::self.step[G3], self.k[G3].y::self.step[G3]])
        # B4 channel
        self.signal.append(img.raw_image[self.k[B4].x::self.step[B4], self.k[B4].y::self.step[B4]])
        # img.close() gives a Segmentation fault, core dumped !!
        # We can't even use a context manager for this
        # img.close()


    def setROI(self, roi):
        if type(roi) == str:
            self.roi = ROI.strproi(roi_str)
        elif type(roi) == ROI:
            self.roi = roi


    def getJulianDate(self):
        jd2000, mjd = jdcal.gcal2jd(self._date.year, self._date.month, self._date.day)
        fraction = (self._date.hour*3600 + self._date.minute*60 + self._date.second)/86400.0
        return jd2000, mjd + fraction


    def night(self):
        '''Observation night as a grouping attribute'''
        jd2000, mjd = self.getJulianDate()
        mjd -= 0.5  # Take it 12 hours before and make sure it is the same YYYY-MM-DD
        year, month, day, fraction = jdcal.jd2gcal(jd2000, mjd)
        return "{0:04d}-{1:02d}-{2:02d}".format(year, month, day)

        
    def stats(self):
        r1_mean, r1_vari = self._region_stats(self.signal[R1], self.roi)
        g2_mean, g2_vari = self._region_stats(self.signal[G2], self.roi)
        g3_mean, g3_vari = self._region_stats(self.signal[G3], self.roi)
        b4_mean, b4_vari = self._region_stats(self.signal[B4], self.roi)
        result = {
            'name'            : self.name,
            'roi'             : str(self.roi),
            'aver_signal_R1'  : r1_mean,
            'vari_signal_R1'  : r1_vari,
            'aver_signal_G2'  : g2_mean,
            'vari_signal_G2'  : g2_vari,
            'aver_signal_G3'  : g3_mean,
            'vari_signal_G3'  : g3_vari,
            'aver_signal_B4'  : b4_mean,
            'vari_signal_B4'  : b4_vari,
        }
        if self.dkroi:
            self._extract_dark()
            self._add_dark_stats(result)
        
        mean  = [r1_mean, g2_mean, g3_mean, b4_mean]
        stdev = [
            round(math.sqrt(r1_vari),1), 
            round(math.sqrt(g2_vari),1), 
            round(math.sqrt(g3_vari),1), 
            round(math.sqrt(b4_vari),1)
        ]
        log.debug("%s: ROI = %s, \u03BC = %s, \u03C3 = %s ", self.name, self.roi, mean, stdev)
        return result

    # ============== #
    # helper methods #
    # ============== #

    def night(self):
        '''Observation night as a grouping attribute'''
        jd2000, mjd = self.getJulianDate()
        mjd -= 0.5  # Take it 12 hours before and make sure it is the same YYYY-MM-DD
        year, month, day, fraction = jdcal.jd2gcal(jd2000, mjd)
        return "{0:04d}-{1:02d}-{2:02d}".format(year, month, day)

    def _fraction_to_float(self, value):
        try:
            value = float(value)
        except TypeError:
            pass    # This handles when value is None
        except ValueError:
            matchobj = FRACTION_REGEXP.search(value)
            if matchobj:
                value = float(matchobj.group(1))/float(matchobj.group(2))
        return value


    def _iso8601(self, tstamp):
        date = None
        for fmt in ["%Y:%m:%d %H:%M:%S", "%Y:%m:%d %H:%M:%S"]:
            try:
                self._date = datetime.datetime.strptime(tstamp, fmt)
            except ValueError:
                continue
            else:
                break
        if not self._date:
            raise TimestampError(tstamp)
        else:
            return self._date.strftime(DEF_TSTAMP)

    def _lookup(self):
        '''
        Load camera configuration from configuration file
        '''

        self.k, self.step = self.cache.lookup(self.model)


    def _region_stats(self, data, region):
        r = data[region.y1:region.y2, region.x1:region.x2]
        return round(r.mean(),1), round(r.var(),3)


    def _center_roi(self):
        '''Conditionally sets the Region of interest around the image center'''
        if self.roi.x1 == 0 and self.roi.y1 == 0:
            width, height = self.roi.dimensions()
            x = np.int(self.signal[G2].shape[1] / 2 - width//2)   # atento: eje X  shape[1]
            y = np.int(self.signal[G2].shape[0] / 2 - height//2)  # atento: eje Y  shape[0]
            self.roi += Point(x,y)  # Shift ROI using this point
        

    def _extract_dark(self):
        self.dark.append(self.signal[R1][-410: , -610:])   # No se de donde salen estos numeros
        self.dark.append(self.signal[G2][-410: , -610:])
        self.dark.append(self.signal[G3][-410: , -610:])
        self.dark.append(self.signal[B4][-410: , -610:])


    def _add_dark_stats(self, mydict):
        r1_aver_dark,   r1_vari_dark   = self._region_stats(self.dark[R1], self.dkroi)
        g2_aver_dark,   g2_vari_dark   = self._region_stats(self.dark[G2], self.dkroi)
        g3_aver_dark,   g3_vari_dark   = self._region_stats(self.dark[G3], self.dkroi)
        b4_aver_dark,   b4_vari_dark   = self._region_stats(self.dark[B4], self.dkroi)
        self.HEADERS.extend([
                'aver_dark_R1', 'vari_dark_R1', 'aver_dark_G2', 'vari_dark_G2',
                'aver_dark_G3', 'vari_dark_G3', 'aver_dark_B4', 'vari_dark_B4'
                ])
        mydict['aver_dark_R1'] = r1_aver_dark
        mydict['vari_dark_R1'] = r1_vari_dark
        mydict['aver_dark_G2'] = g2_aver_dark
        mydict['vari_dark_G2'] = g2_vari_dark
        mydict['aver_dark_G3'] = g3_aver_dark
        mydict['vari_dark_G3'] = g3_vari_dark
        mydict['aver_dark_B4'] = b4_aver_dark
        mydict['vari_dark_B4'] = b4_vari_dark

    def _exiftool(self):
        '''Load EXIF Data using exiftool subprocess'''
        result = subprocess.run(["exiftool", self.filepath],
                                stdout=subprocess.PIPE, check=True)
        lines = result.stdout.decode('utf-8').split('\n')
        # Parse lines using colon as delimiter
        # and updates the self.exif dictionary
        exif_re = re.compile(r'^([^:]+):(.+)')
        for line in lines:
            matchobj = exif_re.search(line)
            if matchobj:
                key   = matchobj.group(1).strip()
                value = matchobj.group(2).strip()
                self.exif[key] = value

        # Emulates the expected EXIF keywords used in the main EXIF library
        self.exif["Image Model"]          = self.exif["Camera Model Name"]
        self.exif["EXIF ISOSpeedRatings"] = self.exif["ISO"]
        self.exif["Image DateTime"]       = self.exif["Date/Time Original"]
        # Focal length requires additional parsing
        focal_re = re.compile(r'^([^ ]+)')
        matchobj = focal_re.search(self.exif["Focal Length"])
        focal = int(float(matchobj.group(1)))
        self.exif["EXIF FocalLength"]     = focal
        self.exif["EXIF FNumber"]         = self.exif["F Number"]
        self.exif["EXIF ExposureTime"]    = self.exif["Exposure Time"]
        CameraImage.ExiftoolFixed = True

       



# This file helps to compute a version number in source trees obtained from
# git-archive tarball (such as those provided by githubs download-from-tag
# feature). Distribution tarballs (built by setup.py sdist) and build
# directories (produced by setup.py build) will contain a much shorter file
# that just contains the computed version number.

# This file is released into the public domain. Generated by
# versioneer-0.18 (https://github.com/warner/python-versioneer)

"""Git implementation of _version.py."""

import errno
import os
import re
import subprocess
import sys


def get_keywords():
    """Get the keywords needed to look up the version information."""
    # these strings will be replaced by git during git-archive.
    # setup.py/versioneer.py will grep for the variable names, so they must
    # each be defined on a line of their own. _version.py will just call
    # get_keywords().
    git_refnames = "$Format:%d$"
    git_full = "$Format:%H$"
    git_date = "$Format:%ci$"
    keywords = {"refnames": git_refnames, "full": git_full, "date": git_date}
    return keywords


class VersioneerConfig:
    """Container for Versioneer configuration parameters."""


def get_config():
    """Create, populate and return the VersioneerConfig() object."""
    # these strings are filled in when 'setup.py versioneer' creates
    # _version.py
    cfg = VersioneerConfig()
    cfg.VCS = "git"
    cfg.style = "pep440"
    cfg.tag_prefix = ""
    cfg.parentdir_prefix = ""
    cfg.versionfile_source = "azotea/_version.py"
    cfg.verbose = False
    return cfg


class NotThisMethod(Exception):
    """Exception raised if a method is not valid for the current scenario."""


LONG_VERSION_PY = {}
HANDLERS = {}


def register_vcs_handler(vcs, method):  # decorator
    """Decorator to mark a method as the handler for a particular VCS."""
    def decorate(f):
        """Store f in HANDLERS[vcs][method]."""
        if vcs not in HANDLERS:
            HANDLERS[vcs] = {}
        HANDLERS[vcs][method] = f
        return f
    return decorate


def run_command(commands, args, cwd=None, verbose=False, hide_stderr=False,
                env=None):
    """Call the given command(s)."""
    assert isinstance(commands, list)
    p = None
    for c in commands:
        try:
            dispcmd = str([c] + args)
            # remember shell=False, so use git.cmd on windows, not just git
            p = subprocess.Popen([c] + args, cwd=cwd, env=env,
                                 stdout=subprocess.PIPE,
                                 stderr=(subprocess.PIPE if hide_stderr
                                         else None))
            break
        except EnvironmentError:
            e = sys.exc_info()[1]
            if e.errno == errno.ENOENT:
                continue
            if verbose:
                print("unable to run %s" % dispcmd)
                print(e)
            return None, None
    else:
        if verbose:
            print("unable to find command, tried %s" % (commands,))
        return None, None
    stdout = p.communicate()[0].strip()
    if sys.version_info[0] >= 3:
        stdout = stdout.decode()
    if p.returncode != 0:
        if verbose:
            print("unable to run %s (error)" % dispcmd)
            print("stdout was %s" % stdout)
        return None, p.returncode
    return stdout, p.returncode


def versions_from_parentdir(parentdir_prefix, root, verbose):
    """Try to determine the version from the parent directory name.

    Source tarballs conventionally unpack into a directory that includes both
    the project name and a version string. We will also support searching up
    two directory levels for an appropriately named parent directory
    """
    rootdirs = []

    for i in range(3):
        dirname = os.path.basename(root)
        if dirname.startswith(parentdir_prefix):
            return {"version": dirname[len(parentdir_prefix):],
                    "full-revisionid": None,
                    "dirty": False, "error": None, "date": None}
        else:
            rootdirs.append(root)
            root = os.path.dirname(root)  # up a level

    if verbose:
        print("Tried directories %s but none started with prefix %s" %
              (str(rootdirs), parentdir_prefix))
    raise NotThisMethod("rootdir doesn't start with parentdir_prefix")


@register_vcs_handler("git", "get_keywords")
def git_get_keywords(versionfile_abs):
    """Extract version information from the given file."""
    # the code embedded in _version.py can just fetch the value of these
    # keywords. When used from setup.py, we don't want to import _version.py,
    # so we do it with a regexp instead. This function is not used from
    # _version.py.
    keywords = {}
    try:
        f = open(versionfile_abs, "r")
        for line in f.readlines():
            if line.strip().startswith("git_refnames ="):
                mo = re.search(r'=\s*"(.*)"', line)
                if mo:
                    keywords["refnames"] = mo.group(1)
            if line.strip().startswith("git_full ="):
                mo = re.search(r'=\s*"(.*)"', line)
                if mo:
                    keywords["full"] = mo.group(1)
            if line.strip().startswith("git_date ="):
                mo = re.search(r'=\s*"(.*)"', line)
                if mo:
                    keywords["date"] = mo.group(1)
        f.close()
    except EnvironmentError:
        pass
    return keywords


@register_vcs_handler("git", "keywords")
def git_versions_from_keywords(keywords, tag_prefix, verbose):
    """Get version information from git keywords."""
    if not keywords:
        raise NotThisMethod("no keywords at all, weird")
    date = keywords.get("date")
    if date is not None:
        # git-2.2.0 added "%cI", which expands to an ISO-8601 -compliant
        # datestamp. However we prefer "%ci" (which expands to an "ISO-8601
        # -like" string, which we must then edit to make compliant), because
        # it's been around since git-1.5.3, and it's too difficult to
        # discover which version we're using, or to work around using an
        # older one.
        date = date.strip().replace(" ", "T", 1).replace(" ", "", 1)
    refnames = keywords["refnames"].strip()
    if refnames.startswith("$Format"):
        if verbose:
            print("keywords are unexpanded, not using")
        raise NotThisMethod("unexpanded keywords, not a git-archive tarball")
    refs = set([r.strip() for r in refnames.strip("()").split(",")])
    # starting in git-1.8.3, tags are listed as "tag: foo-1.0" instead of
    # just "foo-1.0". If we see a "tag: " prefix, prefer those.
    TAG = "tag: "
    tags = set([r[len(TAG):] for r in refs if r.startswith(TAG)])
    if not tags:
        # Either we're using git < 1.8.3, or there really are no tags. We use
        # a heuristic: assume all version tags have a digit. The old git %d
        # expansion behaves like git log --decorate=short and strips out the
        # refs/heads/ and refs/tags/ prefixes that would let us distinguish
        # between branches and tags. By ignoring refnames without digits, we
        # filter out many common branch names like "release" and
        # "stabilization", as well as "HEAD" and "master".
        tags = set([r for r in refs if re.search(r'\d', r)])
        if verbose:
            print("discarding '%s', no digits" % ",".join(refs - tags))
    if verbose:
        print("likely tags: %s" % ",".join(sorted(tags)))
    for ref in sorted(tags):
        # sorting will prefer e.g. "2.0" over "2.0rc1"
        if ref.startswith(tag_prefix):
            r = ref[len(tag_prefix):]
            if verbose:
                print("picking %s" % r)
            return {"version": r,
                    "full-revisionid": keywords["full"].strip(),
                    "dirty": False, "error": None,
                    "date": date}
    # no suitable tags, so version is "0+unknown", but full hex is still there
    if verbose:
        print("no suitable tags, using unknown + full revision id")
    return {"version": "0+unknown",
            "full-revisionid": keywords["full"].strip(),
            "dirty": False, "error": "no suitable tags", "date": None}


@register_vcs_handler("git", "pieces_from_vcs")
def git_pieces_from_vcs(tag_prefix, root, verbose, run_command=run_command):
    """Get version from 'git describe' in the root of the source tree.

    This only gets called if the git-archive 'subst' keywords were *not*
    expanded, and _version.py hasn't already been rewritten with a short
    version string, meaning we're inside a checked out source tree.
    """
    GITS = ["git"]
    if sys.platform == "win32":
        GITS = ["git.cmd", "git.exe"]

    out, rc = run_command(GITS, ["rev-parse", "--git-dir"], cwd=root,
                          hide_stderr=True)
    if rc != 0:
        if verbose:
            print("Directory %s not under git control" % root)
        raise NotThisMethod("'git rev-parse --git-dir' returned error")

    # if there is a tag matching tag_prefix, this yields TAG-NUM-gHEX[-dirty]
    # if there isn't one, this yields HEX[-dirty] (no NUM)
    describe_out, rc = run_command(GITS, ["describe", "--tags", "--dirty",
                                          "--always", "--long",
                                          "--match", "%s*" % tag_prefix],
                                   cwd=root)
    # --long was added in git-1.5.5
    if describe_out is None:
        raise NotThisMethod("'git describe' failed")
    describe_out = describe_out.strip()
    full_out, rc = run_command(GITS, ["rev-parse", "HEAD"], cwd=root)
    if full_out is None:
        raise NotThisMethod("'git rev-parse' failed")
    full_out = full_out.strip()

    pieces = {}
    pieces["long"] = full_out
    pieces["short"] = full_out[:7]  # maybe improved later
    pieces["error"] = None

    # parse describe_out. It will be like TAG-NUM-gHEX[-dirty] or HEX[-dirty]
    # TAG might have hyphens.
    git_describe = describe_out

    # look for -dirty suffix
    dirty = git_describe.endswith("-dirty")
    pieces["dirty"] = dirty
    if dirty:
        git_describe = git_describe[:git_describe.rindex("-dirty")]

    # now we have TAG-NUM-gHEX or HEX

    if "-" in git_describe:
        # TAG-NUM-gHEX
        mo = re.search(r'^(.+)-(\d+)-g([0-9a-f]+)$', git_describe)
        if not mo:
            # unparseable. Maybe git-describe is misbehaving?
            pieces["error"] = ("unable to parse git-describe output: '%s'"
                               % describe_out)
            return pieces

        # tag
        full_tag = mo.group(1)
        if not full_tag.startswith(tag_prefix):
            if verbose:
                fmt = "tag '%s' doesn't start with prefix '%s'"
                print(fmt % (full_tag, tag_prefix))
            pieces["error"] = ("tag '%s' doesn't start with prefix '%s'"
                               % (full_tag, tag_prefix))
            return pieces
        pieces["closest-tag"] = full_tag[len(tag_prefix):]

        # distance: number of commits since tag
        pieces["distance"] = int(mo.group(2))

        # commit: short hex revision ID
        pieces["short"] = mo.group(3)

    else:
        # HEX: no tags
        pieces["closest-tag"] = None
        count_out, rc = run_command(GITS, ["rev-list", "HEAD", "--count"],
                                    cwd=root)
        pieces["distance"] = int(count_out)  # total number of commits

    # commit date: see ISO-8601 comment in git_versions_from_keywords()
    date = run_command(GITS, ["show", "-s", "--format=%ci", "HEAD"],
                       cwd=root)[0].strip()
    pieces["date"] = date.strip().replace(" ", "T", 1).replace(" ", "", 1)

    return pieces


def plus_or_dot(pieces):
    """Return a + if we don't already have one, else return a ."""
    if "+" in pieces.get("closest-tag", ""):
        return "."
    return "+"


def render_pep440(pieces):
    """Build up version string, with post-release "local version identifier".

    Our goal: TAG[+DISTANCE.gHEX[.dirty]] . Note that if you
    get a tagged build and then dirty it, you'll get TAG+0.gHEX.dirty

    Exceptions:
    1: no tags. git_describe was just HEX. 0+untagged.DISTANCE.gHEX[.dirty]
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"] or pieces["dirty"]:
            rendered += plus_or_dot(pieces)
            rendered += "%d.g%s" % (pieces["distance"], pieces["short"])
            if pieces["dirty"]:
                rendered += ".dirty"
    else:
        # exception #1
        rendered = "0+untagged.%d.g%s" % (pieces["distance"],
                                          pieces["short"])
        if pieces["dirty"]:
            rendered += ".dirty"
    return rendered


def render_pep440_pre(pieces):
    """TAG[.post.devDISTANCE] -- No -dirty.

    Exceptions:
    1: no tags. 0.post.devDISTANCE
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"]:
            rendered += ".post.dev%d" % pieces["distance"]
    else:
        # exception #1
        rendered = "0.post.dev%d" % pieces["distance"]
    return rendered


def render_pep440_post(pieces):
    """TAG[.postDISTANCE[.dev0]+gHEX] .

    The ".dev0" means dirty. Note that .dev0 sorts backwards
    (a dirty tree will appear "older" than the corresponding clean one),
    but you shouldn't be releasing software with -dirty anyways.

    Exceptions:
    1: no tags. 0.postDISTANCE[.dev0]
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"] or pieces["dirty"]:
            rendered += ".post%d" % pieces["distance"]
            if pieces["dirty"]:
                rendered += ".dev0"
            rendered += plus_or_dot(pieces)
            rendered += "g%s" % pieces["short"]
    else:
        # exception #1
        rendered = "0.post%d" % pieces["distance"]
        if pieces["dirty"]:
            rendered += ".dev0"
        rendered += "+g%s" % pieces["short"]
    return rendered


def render_pep440_old(pieces):
    """TAG[.postDISTANCE[.dev0]] .

    The ".dev0" means dirty.

    Eexceptions:
    1: no tags. 0.postDISTANCE[.dev0]
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"] or pieces["dirty"]:
            rendered += ".post%d" % pieces["distance"]
            if pieces["dirty"]:
                rendered += ".dev0"
    else:
        # exception #1
        rendered = "0.post%d" % pieces["distance"]
        if pieces["dirty"]:
            rendered += ".dev0"
    return rendered


def render_git_describe(pieces):
    """TAG[-DISTANCE-gHEX][-dirty].

    Like 'git describe --tags --dirty --always'.

    Exceptions:
    1: no tags. HEX[-dirty]  (note: no 'g' prefix)
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        if pieces["distance"]:
            rendered += "-%d-g%s" % (pieces["distance"], pieces["short"])
    else:
        # exception #1
        rendered = pieces["short"]
    if pieces["dirty"]:
        rendered += "-dirty"
    return rendered


def render_git_describe_long(pieces):
    """TAG-DISTANCE-gHEX[-dirty].

    Like 'git describe --tags --dirty --always -long'.
    The distance/hash is unconditional.

    Exceptions:
    1: no tags. HEX[-dirty]  (note: no 'g' prefix)
    """
    if pieces["closest-tag"]:
        rendered = pieces["closest-tag"]
        rendered += "-%d-g%s" % (pieces["distance"], pieces["short"])
    else:
        # exception #1
        rendered = pieces["short"]
    if pieces["dirty"]:
        rendered += "-dirty"
    return rendered


def render(pieces, style):
    """Render the given version pieces into the requested style."""
    if pieces["error"]:
        return {"version": "unknown",
                "full-revisionid": pieces.get("long"),
                "dirty": None,
                "error": pieces["error"],
                "date": None}

    if not style or style == "default":
        style = "pep440"  # the default

    if style == "pep440":
        rendered = render_pep440(pieces)
    elif style == "pep440-pre":
        rendered = render_pep440_pre(pieces)
    elif style == "pep440-post":
        rendered = render_pep440_post(pieces)
    elif style == "pep440-old":
        rendered = render_pep440_old(pieces)
    elif style == "git-describe":
        rendered = render_git_describe(pieces)
    elif style == "git-describe-long":
        rendered = render_git_describe_long(pieces)
    else:
        raise ValueError("unknown style '%s'" % style)

    return {"version": rendered, "full-revisionid": pieces["long"],
            "dirty": pieces["dirty"], "error": None,
            "date": pieces.get("date")}


def get_versions():
    """Get version information or return default if unable to do so."""
    # I am in _version.py, which lives at ROOT/VERSIONFILE_SOURCE. If we have
    # __file__, we can work backwards from there to the root. Some
    # py2exe/bbfreeze/non-CPython implementations don't do __file__, in which
    # case we can only use expanded keywords.

    cfg = get_config()
    verbose = cfg.verbose

    try:
        return git_versions_from_keywords(get_keywords(), cfg.tag_prefix,
                                          verbose)
    except NotThisMethod:
        pass

    try:
        root = os.path.realpath(__file__)
        # versionfile_source is the relative path from the top of the source
        # tree (where the .git directory might live) to this file. Invert
        # this to find the root from __file__.
        for i in cfg.versionfile_source.split('/'):
            root = os.path.dirname(root)
    except NameError:
        return {"version": "0+unknown", "full-revisionid": None,
                "dirty": None,
                "error": "unable to find root of source tree",
                "date": None}

    try:
        pieces = git_pieces_from_vcs(cfg.tag_prefix, root, verbose)
        return render(pieces, cfg.style)
    except NotThisMethod:
        pass

    try:
        if cfg.parentdir_prefix:
            return versions_from_parentdir(cfg.parentdir_prefix, root, verbose)
    except NotThisMethod:
        pass

    return {"version": "0+unknown", "full-revisionid": None,
            "dirty": None,
            "error": "unable to compute version", "date": None}



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------
# Copyright (c) 2020
#
# See the LICENSE file for details
# see the AUTHORS file for authors
# ----------------------------------------------------------------------

#--------------------
# System wide imports
# -------------------

import os.path
import glob
import logging
import shutil

# Python3 catch
try:
    raw_input
except:
    raw_input = input 


# ---------------------
# Third party libraries
# ---------------------

#--------------
# local imports
# -------------

from .      import DEF_DBASE, AZOTEA_BASE_DIR
from .utils import paging

# ----------------
# Module constants
# ----------------

BACKUP_DIR = os.path.join(AZOTEA_BASE_DIR, "backup")

# -----------------------
# Module global variables
# -----------------------

log = logging.getLogger("azotea")

# -----------------------
# Module global functions
# -----------------------


# =====================
# Command esntry points
# =====================


def backup_list(connection, options):
	connection.close()
	file_list = glob.glob(os.path.join(BACKUP_DIR, '*.*' ))
	data = [ [os.path.basename(f)] for f in file_list ]
	headers = ["backup"]
	paging(data, headers, maxsize=len(file_list))


def backup_delete(connection, options):
	connection.close()
	file_path = os.path.join(BACKUP_DIR, options.bak_file)
	os.remove(file_path)
	log.info("Done.")


def backup_restore(connection, options):
	connection.close()
	backup_file = os.path.join(BACKUP_DIR, options.bak_file)
	if not options.non_interactive:
		raw_input("Are you sure ???? <Enter> to continue or [Ctrl-C] to abort")
	shutil.copy2(backup_file, DEF_DBASE)
	log.info("Done.")
	


# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------
# Copyright (c) 2020
#
# See the LICENSE file for details
# see the AUTHORS file for authors
# ----------------------------------------------------------------------

#--------------------
# System wide imports
# -------------------

import sys
import argparse
import sqlite3
import os
import os.path
import logging
import logging.handlers
import traceback
import shutil
import datetime
import hashlib
import zipfile

#--------------
# local imports
# -------------

from .           import *
from .           import __version__
from .exceptions import *
from .config     import load_config_file
from .zenodo     import zenodo_licenses, zenodo_list, zenodo_delete, zenodo_pipeline
# -----------------------
# Module global variables
# -----------------------

log = logging.getLogger("azotenodo")


# -----------------------
# Module global functions
# -----------------------

def configureLogging(options):
	if options.verbose:
		level = logging.DEBUG
	elif options.quiet:
		level = logging.WARN
	else:
		level = logging.INFO
	
	log.setLevel(level)
	# Log formatter
	#fmt = logging.Formatter('%(asctime)s - %(name)s [%(levelname)s] %(message)s')
	fmt = logging.Formatter('%(asctime)s [%(levelname)s] %(message)s')
	# create console handler and set level to debug
	if not options.no_console:
		ch = logging.StreamHandler()
		ch.setFormatter(fmt)
		ch.setLevel(level)
		log.addHandler(ch)
	# Create a file handler
	if options.log_file:
		#fh = logging.handlers.WatchedFileHandler(options.log_file)
		fh = logging.FileHandler(options.log_file)
		fh.setFormatter(fmt)
		fh.setLevel(level)
		log.addHandler(fh)


def python2_warning():
	if sys.version_info[0] < 3:
		log.warning("This software des not run under Python 2 !")


def setup(options):
	
	python2_warning()
	if not os.path.exists(AZOTEA_BASE_DIR):
		log.info("Creating {0} directory".format(AZOTEA_BASE_DIR))
		os.mkdir(AZOTEA_BASE_DIR)
	if not os.path.exists(AZOTEA_CFG_DIR):
		log.info("Creating {0} directory".format(AZOTEA_CFG_DIR))
		os.mkdir(AZOTEA_CFG_DIR)
	if not os.path.exists(AZOTEA_CSV_DIR):
		log.info("Creating {0} directory".format(AZOTEA_CSV_DIR))
		os.mkdir(AZOTEA_CSV_DIR)
	if not os.path.exists(AZOTEA_ZIP_DIR):
		log.info("Creating {0} directory".format(AZOTEA_ZIP_DIR))
		os.mkdir(AZOTEA_ZIP_DIR)
	if not os.path.exists(AZOTEA_LOG_DIR):
		log.info("Creating {0} directory".format(AZOTEA_LOG_DIR))
		os.mkdir(AZOTEA_LOG_DIR)
	if not os.path.exists(options.config):
		shutil.copy2(DEF_CONFIG_TPL, DEF_CONFIG)
		log.info("Created {0} file, please review it".format(DEF_CONFIG))


# =================== #
# THE ARGUMENT PARSER #
# =================== #

def createParser():
	# create the top-level parser
	name = os.path.split(os.path.dirname(sys.argv[0]))[-1]
	parser    = argparse.ArgumentParser(prog=name, description="AZOTEA export to ZENODO tool")

	# Global options
	parser.add_argument('--version', action='version', version='{0} {1}'.format(name, __version__))
	group1 = parser.add_mutually_exclusive_group()
	group1.add_argument('-v', '--verbose', action='store_true', help='Verbose output.')
	group1.add_argument('-q', '--quiet',   action='store_true', help='Quiet output.')
	parser.add_argument('-nk','--no-console', action='store_true', help='Do not log to console.')
	parser.add_argument('--log-file', type=str, default=None, help='Optional log file')
	parser.add_argument('--config', type=str, default=DEF_CONFIG, help='Optional alternate configuration file')
	parser.add_argument('-t', '--test',   action='store_true', help='Use the ZENODO Sandbox environment')


	subparser = parser.add_subparsers(dest='command')

	parser_list     = subparser.add_parser('list',      help='list contents')
	parser_licenses = subparser.add_parser('licenses',  help='list Zenodo available publication licenses')
	parser_delete   = subparser.add_parser('delete',    help='delete draft Zenodo depositions')
	parser_pipeline = subparser.add_parser('pipeline',  help='The full pipeline (deposit/upload/publish) with versioning')
	

	# -----------
	# List action
	# -----------
	parser_list.add_argument('--published', action='store_true', default=False, help='List published results only')
	parser_list.add_argument('--title', type=str, default=None,  help='Search by title')

	# -------------
	# Delete action
	# -------------
	parser_delete.add_argument('--id',type=int, required=True,  help='Zenodo ID to delete')

	# --------------
	# Pipeline action
	# --------------

	parser_pipeline.add_argument('--title',    type=str, nargs='+', default=AZOTEA_PUBL_TITLE, help='Optional Publication Title')
	parser_pipeline.add_argument('--csv-dir',  type=str, default=AZOTEA_CSV_DIR,    help='Optional CSV file dir')
	parser_pipeline.add_argument('--zip-file' ,type=str, default=DEF_ZIP_FILE,      help='Optional ZIP File to create with all CSV files')
	parser_pipeline.add_argument('--community',type=str, default=AZOTEA_COMMUNITY,  help='Optional community where to publish the dataset')
	parser_pipeline.add_argument('--version',  type=str, default=None,              help='Optional version string to tag, useful for tests')
	
	return parser




# ================ #
# MAIN ENTRY POINT #
# ================ #

def main():
	'''
	Utility entry point
	'''
	try:
		options = createParser().parse_args(sys.argv[1:])
		configureLogging(options)
		command  = options.command
		log.info("============== AZOTENODO {0} ==============".format(__version__))
		file_options = load_config_file(options.config)
		file_options = argparse.Namespace(**file_options)
		setup(options)
		globals()['zenodo_' + command](options, file_options)
	except KeyboardInterrupt as e:
		log.critical("[%s] Interrupted by user ", __name__)
	except Exception as e:
		log.critical("[%s] Fatal error => %s", __name__, str(e) )
		traceback.print_exc()
	finally:
		pass

main()



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------
# Copyright (c) 2020
#
# See the LICENSE file for details
# see the AUTHORS file for authors
# ----------------------------------------------------------------------

#--------------------
# System wide imports
# -------------------

import logging

# ---------------------
# Third party libraries
# ---------------------

#--------------
# local imports
# -------------

from .utils      import paging

# ----------------
# Module constants
# ----------------


# -----------------------
# Module global variables
# -----------------------

# =======================
# Module global functions
# =======================

# -----------------
# Utility functions
# -----------------

def lookup_session(connection, session):
	'''Get one  session'''
	row = {'session': session}
	cursor = connection.cursor()
	cursor.execute('''
		SELECT session
		FROM image_t
		WHERE session = session 
		''', row)
	return cursor.fetchone()[0]


def latest_session(connection):
	'''Get Last recorded session'''
	cursor = connection.cursor()
	cursor.execute('''
		SELECT MAX(session)
		FROM image_t 
		''')
	return cursor.fetchone()[0]


def session_all_count(cursor):
	cursor.execute(
		'''
		SELECT COUNT(*)
		FROM image_t
		GROUP BY session
		''')
	result = [ x[0] for x in cursor.fetchall()]
	return sum(result)


def session_session_count(cursor, session):
	row = {'session': session}
	cursor.execute(
		'''
		SELECT COUNT(*)
		FROM image_t
		WHERE session = :session
		GROUP BY session, type, state
		''',row)
	result = [ x[0] for x in cursor.fetchall()]
	return sum(result)


# ------------------
# Database iterables
# ------------------


def session_summary_all_iterable(connection, session):
	cursor = connection.cursor()
	count = session_all_count(cursor)
	cursor.execute(
		'''
		SELECT session, type, s.label, COUNT(*)
		FROM image_t
		JOIN state_t AS s USING(state)
		GROUP BY session, type, state
		ORDER BY session DESC, type, state 
		''')
	return cursor, count


def session_extended_all_iterable(connection, session):
	cursor = connection.cursor()
	count = session_all_count(cursor)
	cursor.execute(
		'''
		SELECT session, name, tstamp, type, s.label
		FROM image_t
		JOIN state_t AS s USING(state)
		ORDER BY session DESC, name ASC, type
		''')
	return cursor, count


def session_summary_session_iterable(connection, session):
	row = {'session': session}
	cursor = connection.cursor()
	count = session_session_count(cursor, session)
	cursor.execute(
		'''
		SELECT session, type, s.label, COUNT(*)
		FROM image_t
		JOIN state_t AS s USING(state)
		WHERE session = :session
		GROUP BY session, type, state
		ORDER BY session DESC, type, state 
		''', row)
	return cursor, count


def session_extended_session_iterable(connection, session):
	row = {'session': session}
	cursor = connection.cursor()
	count = session_session_count(cursor, session)
	cursor.execute(
		'''
		SELECT session, name, tstamp, type, s.label
		FROM image_t
		JOIN state_t AS s USING(state)
		WHERE session = :session
		ORDER BY session DESC, name ASC, type
		''', row)
	return cursor, count


# ------------------
# Database inserters
# ------------------



# ==================================
# Image View sumcommands and options
# ==================================


SUMMARY_HEADERS = [
	'Session',
	'Type',
	'State',
	'# Images',
]


EXTENDED_HEADERS = [
	'Session',
	'Name',
	'Date',
	'Type',
	'State',
]




def do_session_view(connection, session, iterable, headers, options):
	cursor, count = iterable(connection, session)
	paging(cursor, headers, maxsize=count, page_size=options.page_size)


# =====================
# Command esntry points
# =====================


def session_current(connection, options):
	session = latest_session(connection)
	if options.extended:
		headers = EXTENDED_HEADERS
		iterable = session_extended_session_iterable 
	else:
		headers = SUMMARY_HEADERS
		iterable = session_summary_session_iterable
	do_session_view(connection, session, iterable, headers, options)


def session_list(connection, options):
	if options.all and options.extended:
		headers = EXTENDED_HEADERS
		session = None
		iterable = session_extended_all_iterable
	elif options.all and not options.extended:
		headers = SUMMARY_HEADERS
		session = None
		iterable = session_summary_all_iterable
	elif not options.all and options.extended:
		headers = EXTENDED_HEADERS
		session = lookup_session(connection, options.session)
		iterable = session_extended_session_iterable
	else:
		headers = SUMMARY_HEADERS
		session = lookup_session(connection, options.session)
		iterable = session_summary_session_iterable
	do_session_view(connection, session, iterable, headers, options)





# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------
# Copyright (c) 2020
#
# See the LICENSE file for details
# see the AUTHORS file for authors
# ----------------------------------------------------------------------

#--------------------
# System wide imports
# -------------------

import sys
import sqlite3
import os.path
import glob
import logging
import csv
import datetime
import math
import hashlib
import time
import re
import collections

# ---------------------
# Third party libraries
# ---------------------

#--------------
# local imports
# -------------

from .           import AZOTEA_CSV_DIR, AZOTEA_CFG_DIR
from .camera     import CameraImage, CameraCache, MetadataError, ConfigError
from .utils      import merge_two_dicts, paging, LogCounter
from .exceptions import MixingCandidates, NoUserInfoError
from .config     import load_config_file, merge_options


# ----------------
# Module constants
# ----------------

# values for the 'state' column in table
REGISTERED       = 0
STATS_COMPUTED   = 1
METADATA_UPDATED = 3
DARK_SUBSTRACTED = 3

# (bit) flags for the 'meta_changes' column in table
NO_CHANGES       = 0
CAMERA_CHANGES   = 1
OBSERVER_CHANGES = 2

# Values for the 'tyoe' column
LIGHT_FRAME = "LIGHT"
BIAS_FRAME  = "BIAS"
DARK_FRAME  = "DARK"
UNKNOWN     = "UNKNOWN"

N_COUNT = 50

# -----------------------
# Module global variables
# -----------------------

log = logging.getLogger("azotea")


if sys.version_info[0] == 2:
	import errno
	class FileExistsError(OSError):
		def __init__(self, msg):
			super(FileExistsError, self).__init__(errno.EEXIST, msg)

# =======================
# Module global functions
# =======================

# -----------------
# Utility functions
# -----------------

def classification_algorithm1(name,  file_path, options):
	if name.upper().startswith(DARK_FRAME):
		result = {'name': name, 'type': DARK_FRAME}
	else:
		result = {'name': name, 'type': LIGHT_FRAME}
	return result


RE_DARK = re.compile(r'.*DARK.*\..{3}')
def classification_algorithm2(name,  file_path, options):
	if RE_DARK.search(name.upper()):
		result = {'name': name, 'type': DARK_FRAME}
	else:
		result = {'name': name, 'type': LIGHT_FRAME}
	return result


def session_processed(connection, session):
	row = {'session': session, 'state': STATS_COMPUTED}
	cursor = connection.cursor()
	cursor.execute('''
		SELECT COUNT(*) 
		FROM image_t 
		WHERE state >= :state
		AND session = :session
		''',row)
	return cursor.fetchone()[0]


def latest_session(connection):
	'''Get Last recorded session'''
	cursor = connection.cursor()
	cursor.execute('''
		SELECT MAX(session)
		FROM image_t 
		''')
	return cursor.fetchone()[0]


def myopen(name, *args):
	if sys.version_info[0] < 3:
		return open(name, *args)
	else:
		return open(name,  *args, newline='')


def hash(filepath):
	'''Compute a hash from the image'''
	BLOCK_SIZE = 1048576 # 1MByte, the size of each read from the file
	
	# md5() was the fastest algorithm I've tried
	# but I detected a collision, so I now use blake2b with twice the digest size
	
	file_hash = hashlib.blake2b(digest_size=32)
	#file_hash = hashlib.md5()
	with open(filepath, 'rb') as f:
		block = f.read(BLOCK_SIZE) 
		while len(block) > 0:
			file_hash.update(block)
			block = f.read(BLOCK_SIZE)
	return file_hash.digest()


def find_by_hash(connection, hash):
	row = {'hash': hash}
	cursor = connection.cursor()
	cursor.execute('''
		SELECT name
		FROM image_t 
		WHERE hash = :hash
		''',row)
	return cursor.fetchone()


def image_session_state_reset(connection, session):
	row = {'session': session, 'state': STATS_COMPUTED, 'new_state': REGISTERED, 'type': UNKNOWN}
	cursor = connection.cursor()
	cursor.execute(
		'''
		UPDATE image_t
		SET state = :new_state, type = :type
		WHERE state >= :state
		AND session = :session
		''', row)
	connection.commit()



def detect_dupl_hashes(names_hashes_list):
	hashes_list = [ item['hash'] for item in names_hashes_list ]
	c = collections.Counter(hashes_list)
	for hsh, count in c.most_common():
		if count > 1:
			for item in names_hashes_list:
				if item['hash'] == hsh:
					name = item['name']
					log.error("Image {0} has a clashing hash {1} ".format(name, hsh))


def register_slow_candidates(connection, names_hashes_list):
	cursor = connection.cursor()
	for item in names_hashes_list:
		try:
			cursor.execute("INSERT INTO candidate_t (name,hash) VALUES (:name,:hash)", item)
		except sqlite3.IntegrityError as e:
			connection.rollback()
			log.warning("Ignoring duplicate candidate in the same dir => %s", item['name'])
		else:
			connection.commit()


def work_dir_to_session(connection, work_dir, filt):
	file_list  = sorted(glob.glob(os.path.join(work_dir, filt)))
	log.info("Found {0} candidates matching filter {1}.".format(len(file_list), filt))
	log.info("Computing hashes. This may take a while")
	names_hashes_list = [ {'name': os.path.basename(p), 'hash': hash(p)} for p in file_list ]
	detect_dupl_hashes(names_hashes_list)
	cursor = connection.cursor()
	cursor.execute("CREATE TEMP TABLE candidate_t (name TEXT, hash BLOB, PRIMARY KEY(hash))")
	try:
		cursor.executemany("INSERT INTO candidate_t (name,hash) VALUES (:name,:hash)", names_hashes_list)
	except sqlite3.IntegrityError as e:
		connection.rollback()
		register_slow_candidates(connection, names_hashes_list)
	else:
		connection.commit()
	# Common images to database and work-dir
	cursor.execute(
		'''
		SELECT MAX(session), MAX(session) == MIN(session)
		FROM image_t
		WHERE hash IN (SELECT hash FROM candidate_t)
		'''
		)
	result = cursor.fetchall()
	if result:
		sessiones, flags = zip(*result)
		if flags[0] and not all(flags):
			raise MixingCandidates(names_common)
		session = sessiones[0]
	else:
		session = None
	return session


def work_dir_cleanup(connection):
	cursor = connection.cursor()
	cursor.execute("DROP TABLE IF EXISTS candidate_t")
	connection.commit()


def master_dark_for(connection, session):
	row = {'session': session}
	cursor = connection.cursor()
	cursor.execute('''
		SELECT COUNT(*) 
		FROM master_dark_t
		WHERE session = :session
		''',row)
	return cursor.fetchone()[0]



# --------------
# Image Register
# --------------

def register_insert_image(connection, row):
	'''slow version to find out the exact duplicate'''
	cursor = connection.cursor()
	cursor.execute(
			'''
			INSERT OR IGNORE INTO image_t (
				name, 
				hash,
				session, 
				type,
				state,
				meta_changes
			) VALUES (
				:name, 
				:hash,
				:session,
				:type,
				:state,
				:changes
			)
			''', row)
	connection.commit()
	

def register_insert_images(connection, rows):
	'''fast version'''
	cursor = connection.cursor()
	cursor.executemany(
			'''
			INSERT OR IGNORE INTO image_t (
				name, 
				hash,
				session,
				type,
				state,
				meta_changes
			) VALUES (
				:name, 
				:hash,
				:session, 
				:type,
				:state,
				:changes
			)
			''', rows)
	connection.commit()
	log.info("Registered %d / %d images in database", cursor.rowcount, len(rows))


def candidates(connection, work_dir, filt, session):
	'''candidate list of images to be inserted/removed to/from the database'''
	# New Images in the work dir that should be added to database
	cursor = connection.cursor()
	# This commented query may take long to execute if the database is large
	# better we leave it fail in the insertion, where the hash duplication is detected
	# cursor.execute(
	# 	'''
	# 	SELECT name, hash
	# 	FROM candidate_t
	# 	WHERE hash NOT IN (SELECT hash FROM image_t)
	# 	'''
	# 	)

	# This query will have far less elements to fetch
	# This will introduce duplicates which will be rejected by the INSERT or IGNORE
	# when inserting new images
	cursor.execute("SELECT name, hash FROM candidate_t")
	result = cursor.fetchall()
	if result:
		#names_to_add, = zip(*result)
		names_to_add = result
	else:
		names_to_add = []
	
	# Images no longer in the work dir, they should be deleted from database
	row = {'session': session}
	cursor.execute(
		'''
		SELECT name, hash
		FROM image_t
		WHERE session = :session
		AND hash NOT IN (SELECT hash FROM candidate_t)
		''', row)
	result = cursor.fetchall()
	if result:
		#names_to_del, = zip(*result)
		names_to_del = result
	else:
		names_to_del = []
	return names_to_add, names_to_del


def register_delete_images(connection, rows):
	'''delete images'''
	cursor = connection.cursor()
	cursor.executemany(
			'''
			DELETE FROM image_t 
			WHERE hash  == :hash
			''', rows)
	connection.commit()
	log.info("Deleted %d / %d images from database", cursor.rowcount, len(rows))


def register_slow(connection, work_dir, names_list, session):
	counter = LogCounter(N_COUNT)
	for name, hsh in names_list:
		file_path = os.path.join(work_dir, name)
		row  = {'name': name, 'hash': hsh, 'session': session, 
		'state': REGISTERED, 'type': UNKNOWN, 'changes': CAMERA_CHANGES + OBSERVER_CHANGES}
		try:
			register_insert_image(connection, row)
		except sqlite3.IntegrityError as e:
			connection.rollback()
			name2, = find_by_hash(connection, row['hash'])
			log.warning("Duplicate => %s EQUALS %s", file_path, name2)
		else:
			counter.tick("Registered %03d images in database (slow method)",level=logging.DEBUG)
			log.debug("%s registered in database", row['name'])
	counter.end("Registered %03d images in database (slow method)")


def register_fast(connection, work_dir, names_list, session):
	rows = []
	for name, hsh in names_list:
		file_path = os.path.join(work_dir, name)
		row  = {'name': name, 'session': session, 
		'state': REGISTERED, 'type': UNKNOWN, 'changes': CAMERA_CHANGES + OBSERVER_CHANGES}
		row['hash'] = hsh
		rows.append(row)
		log.debug("Image %s being registered in database", row['name'])
	register_insert_images(connection, rows)
	

def register_unregister(connection, names_list, session):
	rows = []
	log.info("Unregistering images from database")
	for name, hsh in names_list:
		rows.append({'session': session, 'name': name, 'hash': hsh})
		log.info("Image %s being removed from database", name)
	register_delete_images(connection, rows)
	

def register_log_kept(connection, session):
	# Images  in the work dir already existing in the database
	ARBITRARY_NUMBER = 10
	cursor = connection.cursor()
	row = {'session': session, 'count': ARBITRARY_NUMBER}
	cursor.execute(
		'''
		SELECT COUNT(*)
		FROM image_t
		WHERE hash IN (SELECT hash FROM candidate_t)
		''', row)
	count = cursor.fetchone()[0]
	if count:
		cursor.execute(
			'''
			SELECT name
			FROM image_t
			WHERE hash IN (SELECT hash FROM candidate_t)
			ORDER BY name ASC
			LIMIT :count
			''', row)
		for name, in cursor:
			log.info("Image %s being kept in database", name)
		if count > ARBITRARY_NUMBER:
			log.info("And %d more images being kept in database", count - ARBITRARY_NUMBER)


# Tal como esta montado ahora candidates(), es imposible introducir una imagen
# duplicada porque se cumprueba primero que su hash no esta ya en la BD
# Y por tanbo register_low() es innecesario.
# Sin embargo candidates() podría enlentecerse al aumentar el número de imagenes de la BD
# Por lo que al final register_slow() podría ser una opcion
def do_register(connection, work_dir, filt, session):
	register_deleted = False
	names_to_add, names_to_del = candidates(connection, work_dir, filt, session)
	if names_to_del:
		register_unregister(connection, names_to_del, session)  
		register_log_kept(connection, session)
		register_deleted = True
	if names_to_add:
		register_fast(connection, work_dir, names_to_add, session)
	return register_deleted



# --------------
# Image Classify
# --------------


def classify_update_db(connection, rows):
	cursor = connection.cursor()
	cursor.executemany(
		'''
		UPDATE image_t
		SET type  = :type
		WHERE name = :name
		''', rows)
	connection.commit()


def classify_session_iterable(connection, session):
	row = {'session': session, 'type': UNKNOWN}
	cursor = connection.cursor()
	cursor.execute(
		'''
		SELECT name
		FROM image_t
		WHERE type = :type
		AND session = :session
		''', row)
	return cursor

def classify_log_type(connection, session):
	row = {'session': session}
	cursor = connection.cursor()
	cursor.execute(
		'''
		SELECT type, COUNT(*)
		FROM image_t
		WHERE session = :session
		GROUP BY type
		''', row)
	for imagetyp, n in cursor:
		log.info("% -5s frames: % 3d", imagetyp, n)


def do_classify(connection, session, work_dir, options):
	rows = []
	counter = LogCounter(N_COUNT)
	log.debug("Classifying images")
	for name, in classify_session_iterable(connection, session):
		file_path = os.path.join(work_dir, name)
		row = classification_algorithm2(name, file_path, options)
		log.debug("%s is type %s", name, row['type'])
		counter.tick("Classified %03d images",level=logging.DEBUG)
		rows.append(row)
	if rows:
		classify_update_db(connection, rows)
		counter.end("Classified %03d images")
		classify_log_type(connection, session)
	else:
		log.info("No image type classification is needed")


# -----------
# Image Stats
# -----------

def stats_update_db(connection, rows):
	cursor = connection.cursor()
	cursor.executemany(
		'''
		UPDATE image_t
		SET 
			state               = :state,
			roi                 = :roi, 
			model               = :model,        -- EXIF
			iso                 = :iso,          -- EXIF
			tstamp              = :tstamp,       -- EXIF
			night               = :night,        -- computed from EXIF
			exptime             = :exptime,      -- EXIF
			focal_length        = :focal_length, -- EXIF
			f_number            = :f_number,     -- EXIF
			bias                = :bias,         -- EXIF
			aver_signal_R1  = :aver_signal_R1, 
			aver_signal_G2  = :aver_signal_G2, 
			aver_signal_G3  = :aver_signal_G3,
			aver_signal_B4  = :aver_signal_B4,
			vari_signal_R1  = :vari_signal_R1,
			vari_signal_G2  = :vari_signal_G2,
			vari_signal_G3  = :vari_signal_G3,
			vari_signal_B4  = :vari_signal_B4 
		WHERE hash = :hash
		''', rows)
	connection.commit()


def stats_session_iterable(connection, session):
	row = {'session': session, 'state': STATS_COMPUTED}
	cursor = connection.cursor()
	cursor.execute(
		'''
		SELECT name, hash
		FROM image_t
		WHERE state < :state
		AND session = :session
		ORDER BY name ASC
		''', row)
	return cursor


def stats_unregister(connection, rows):
	'''Unregister an image who gave an exception reaing the pixel data'''
	cursor = connection.cursor()
	cursor.executemany(
			'''
			DELETE FROM image_t 
			WHERE hash  == :hash
			''', rows)
	connection.commit()
	log.info("Deleted %d / %d images from database", cursor.rowcount, len(rows))
	

def do_stats(connection, session, work_dir, options):
	stats_computed_flag = False
	camera_cache = CameraCache(options.camera)
	rows = []
	rows_to_unregister = []
	counter = LogCounter(N_COUNT)
	CameraImage.ExiftoolFixed = False
	log.debug("Computing image statistics")
	for name, hsh in stats_session_iterable(connection, session):
		file_path = os.path.join(work_dir, name)
		image = CameraImage(file_path, camera_cache)
		image.setROI(options.roi)
		counter.tick("Statistics for %03d images done")
		metadata = image.loadEXIF()
		try:
			image.read()
		except Exception as e:
			log.warn("%s: Error while reading pixels", name)
			rows_to_unregister.append({'name': name, 'hash': hsh, 'session':session})
		else:
			row = image.stats()
			row['hash']         = hsh
			row['state']        = STATS_COMPUTED
			row['model']        = metadata['model']
			row['iso']          = metadata['iso']
			row['tstamp']       = metadata['tstamp']
			row['night']        = metadata['night']
			row['exptime']      = metadata['exptime']
			row['bias']         = metadata['bias']
			row['focal_length'] = metadata['focal_length']
			row['f_number']     = metadata['f_number']
			rows.append(row)
	if rows_to_unregister:
		stats_unregister(connection, rows_to_unregister)
	if rows:
		counter.end("Statistics for %03d images done")
		stats_update_db(connection, rows)
		stats_computed_flag = True
	else:
		log.info("No image statistics to be computed")
	if CameraImage.ExiftoolFixed:
		log.warning("exiftool was used to read EXIF metadata")
	return stats_computed_flag


# ---------------
# Metadata Update
# ---------------

def do_metadata(connection, session, options):
	log.debug("Updating Global Metadata")
	row = vars(options)
	row['state']   = METADATA_UPDATED
	row['session'] = session
	row['changes'] = NO_CHANGES
	cursor = connection.cursor()
	cursor.execute('''
			SELECT meta_changes
			FROM image_t
			WHERE session = :session
			LIMIT 1
			''', row)
	flags = cursor.fetchone()
	if flags is None:
		log.info("No image metadata was updated")
		metadata_updated_flag = False
		return metadata_updated_flag
	
	flags = flags[0]
	# Conditionally changes observer and location if given by an event
	if flags & OBSERVER_CHANGES:
		log.info("Updating metadata observer: %s, %s, %s", options.observer, options.organization, options.location)
		cursor.execute('''
		UPDATE image_t 
		SET 
			observer     = :observer,
			obs_family_name = :obs_family_name,
			obs_surname  = :obs_surname,
			organization = :organization,
			email        = :email,
			location     = :location
		WHERE session = :session
		AND   state   < :state
			''', row)

	# Conditionally changes focal lengthn if given by an event
	if flags & CAMERA_CHANGES and options.focal_length is not None:
		log.info("Updating focal length metadata to %d", options.focal_length)
		cursor.execute('''
			UPDATE image_t 
			SET focal_length = :focal_length
			WHERE session = :session
			AND   state   < :state
			''', row)

	# Conditionally changes focal lengthn if given by an event
	if flags & CAMERA_CHANGES and options.f_number is not None:
		log.info("Updating f/ number metadata to %d", options.f_number )
		cursor.execute('''
			UPDATE image_t 
			SET   f_number  = :f_number
			WHERE session = :session
			AND   state   < :state
			''', row)

	# Conditionally chhanges observer and location if given by an event
	if flags & CAMERA_CHANGES and options.bias is not None:
		log.info("Updating bias metadata to %d", options.bias)
		cursor.execute('''
			UPDATE image_t SET bias = :bias
			WHERE session = :session
			AND   state   < :state
			''', row)
	
	# Update state and clerar change flags
	cursor.execute('''
		UPDATE image_t 
		SET 
			state        = :state,
			meta_changes = :changes
		WHERE session = :session
		AND   state   < :state
			''', row)
	connection.commit()
	
	if cursor.rowcount > 0:
		log.info("Updated Global Metadata for %03d images", cursor.rowcount)
		metadata_updated_flag = True
	else:
		log.info("No image metadata was updated")
		metadata_updated_flag = False
	return metadata_updated_flag


# -----------------------------
# Image Apply Dark Substraction
# -----------------------------


def master_dark_db_update_all(connection, session):
	row = {'type': DARK_FRAME, 'state': STATS_COMPUTED}
	cursor = connection.cursor()
	cursor.execute(
		'''
		INSERT OR REPLACE INTO master_dark_t (
			session, 
			roi, 
			N, 
			aver_R1, 
			aver_G2, 
			aver_G3, 
			aver_B4,
			vari_R1, 
			vari_G2, 
			vari_G3, 
			vari_B4,
			min_exptime,
			max_exptime
		)
		SELECT 
			session, 
			MIN(roi), 
			COUNT(*), 
			-- This should be the proper treatment
			-- AVG(aver_signal_R1 - bias),
			-- AVG(aver_signal_G2 - bias),
			-- AVG(aver_signal_G3 - bias),
			-- AVG(aver_signal_B4 - bias),
			-- Current treatment: Do not substract BIAS from raw signal to perform master dark
			AVG(aver_signal_R1),
			AVG(aver_signal_G2),
			AVG(aver_signal_G3),
			AVG(aver_signal_B4),
			SUM(vari_signal_R1)/COUNT(*),
			SUM(vari_signal_G2)/COUNT(*),
			SUM(vari_signal_G3)/COUNT(*),
			SUM(vari_signal_B4)/COUNT(*),
			MIN(exptime),
			MAX(exptime)
		FROM image_t
		WHERE type = :type
		AND   state >= :state
		GROUP BY session
		''', row)
	connection.commit()


def master_dark_db_update_session(connection, session):
	row = {'type': DARK_FRAME, 'state': STATS_COMPUTED, 'session': session}
	cursor = connection.cursor()
	cursor.execute(
		'''
		INSERT OR REPLACE INTO master_dark_t (
			session, 
			roi, 
			N, 
			aver_R1, 
			aver_G2, 
			aver_G3, 
			aver_B4,
			vari_R1, 
			vari_G2, 
			vari_G3, 
			vari_B4,
			min_exptime,
			max_exptime
		)
		SELECT 
			session, 
			MIN(roi), 
			COUNT(*), 
			-- This should be the proper treatment
			-- AVG(aver_signal_R1 - bias),
			-- AVG(aver_signal_G2 - bias),
			-- AVG(aver_signal_G3 - bias),
			-- AVG(aver_signal_B4 - bias),
			-- Current treatment: Do not substract BIAS from raw signal to perform master dark
			AVG(aver_signal_R1),
			AVG(aver_signal_G2),
			AVG(aver_signal_G3),
			AVG(aver_signal_B4),
			SUM(vari_signal_R1)/COUNT(*),
			SUM(vari_signal_G2)/COUNT(*),
			SUM(vari_signal_G3)/COUNT(*),
			SUM(vari_signal_B4)/COUNT(*),
			MIN(exptime),
			MAX(exptime)
		FROM image_t
		WHERE session = :session 
		AND   type    = :type
		AND   state  >= :state
		GROUP BY session
		''', row)
	connection.commit()


def dark_update_columns(connection, session):
	row = {'type': LIGHT_FRAME, 'session': session, 'state': STATS_COMPUTED, 'new_state': DARK_SUBSTRACTED}
	cursor = connection.cursor()
	cursor.execute(
		'''
		UPDATE image_t
		SET
			state        = :new_state,
			aver_dark_R1 = (SELECT aver_R1 FROM master_dark_t WHERE session = :session),
			aver_dark_G2 = (SELECT aver_G2 FROM master_dark_t WHERE session = :session),
			aver_dark_G3 = (SELECT aver_G3 FROM master_dark_t WHERE session = :session),
			aver_dark_B4 = (SELECT aver_B4 FROM master_dark_t WHERE session = :session),
			vari_dark_R1 = (SELECT vari_R1 FROM master_dark_t WHERE session = :session),
			vari_dark_G2 = (SELECT vari_G2 FROM master_dark_t WHERE session = :session),
			vari_dark_G3 = (SELECT vari_G3 FROM master_dark_t WHERE session = :session),
			vari_dark_B4 = (SELECT vari_B4 FROM master_dark_t WHERE session = :session)
		WHERE session = :session
		AND   state BETWEEN :state AND :new_state
		AND   type  = :type
		''',row)
	connection.commit()


def do_apply_dark(connection, session, options):
	master_dark_db_update_session(connection, session)
	if master_dark_for(connection, session):
		log.info("Applying dark substraction to current working directory")
		dark_update_columns(connection, session)
	else:
		log.info("No dark frame found for current working directory")

# -----------
# Image Export
# -----------


EXPORT_HEADERS = [
			'tstamp'         ,
			'name'           ,
			'model'          ,
			'iso'            ,
			'roi'            ,
			'dark_roi'       ,
			'exptime'        ,
			'aver_signal_R1' ,
			'std_signal_R1'  ,
			'aver_signal_G2' ,
			'std_signal_G2'  ,
			'aver_signal_G3' ,
			'std_signal_G3'  ,
			'aver_signal_B4' ,
			'std_signal_B4'  ,
			'aver_dark_R1'   ,
			'std_dark_R1'    ,
			'aver_dark_G2'   ,
			'std_dark_G2'    ,
			'aver_dark_G3'   ,
			'std_dark_G3'    ,
			'aver_dark_B4'   ,
			'std_dark_B4'    ,
			'bias'           ,
		]


def night_iterable(connection, session):
	row = {'session': session}
	cursor = connection.cursor()
	cursor.execute(
		'''
		SELECT  DISTINCT night
		FROM image_t
		WHERE session == :session
		''', row)
	return cursor


# we are not using the image_v VIEW for the time being
# We display the RAW data without dark and bias substraction
def export_session_iterable(connection, session, night):
	row = {'session': session, 'state': STATS_COMPUTED, 'type': LIGHT_FRAME, 'night': night}
	cursor = connection.cursor()
	cursor.execute(
		'''
		SELECT  session,
				observer,
				organization,
				location,
				type,
				tstamp, 
				name, 
				model, 
				iso, 
				roi,
				dark_roi,
				exptime, 
				aver_signal_R1, 
				vari_signal_R1, -- Array index 13
				aver_signal_G2, 
				vari_signal_G2, -- Array index 15
				aver_signal_G3, 
				vari_signal_G3, -- Array index 17
				aver_signal_B4, 
				vari_signal_B4, -- Array index 19
				aver_dark_R1, 
				vari_dark_R1,   -- Array index 21
				aver_dark_G2, 
				vari_dark_G2,   -- Array index 23
				aver_dark_G3, 
				vari_dark_G3,   -- Array index 25
				aver_dark_B4, 
				vari_dark_B4,   -- Array index 27
				bias
		FROM image_t
		WHERE state  >= :state
		AND   type    = :type
		AND   session = :session
		AND   night   = :night
		ORDER BY tstamp ASC
		''', row)
	return cursor

# we are not using the image_v VIEW for the time being
# We display the RAW data without dark and bias substraction
def export_all_iterable(connection):
	row = {'state': STATS_COMPUTED, 'type': LIGHT_FRAME}
	cursor = connection.cursor()
	cursor.execute(
		'''
	   SELECT   session,
				observer,
				organization,
				location,
				type,
				tstamp, 
				name, 
				model, 
				iso, 
				roi,
				dark_roi,
				exptime, 
				aver_signal_R1, 
				vari_signal_R1, -- Array index 13
				aver_signal_G2, 
				vari_signal_G2, -- Array index 15
				aver_signal_G3, 
				vari_signal_G3, -- Array index 17
				aver_signal_B4, 
				vari_signal_B4, -- Array index 19
				aver_dark_R1, 
				vari_dark_R1,   -- Array index 21
				aver_dark_G2, 
				vari_dark_G2,   -- Array index 23
				aver_dark_G3, 
				vari_dark_G3,   -- Array index 25
				aver_dark_B4, 
				vari_dark_B4,   -- Array index 27
				bias
		FROM image_t
		WHERE state >= :state
		AND   type = :type
		ORDER BY observer ASC, tstamp ASC
		''', row)
	return cursor


def var2std(item):
	'''From Variance to StdDev in several columns'''
	index, value = item
	# Calculate stddev from variance and round to one decimal place
	if  index in  [13, 15, 17, 19, 21, 23, 25, 27]:
		value = round(math.sqrt(value),1)
	# Round the aver_signal channels too
	elif index in [12, 14, 16, 18, 20, 22, 24, 26]:
		value = round(value, 1)
	return value



def get_file_path(connection, night, work_dir, options):
	# This is for automatic reductions mainly
	key, ext  = os.path.splitext(options.config)
	key       = os.path.basename(key)
	#wdtag     = os.path.basename(work_dir)
	filename  = "-".join([key, night + '.csv'])
	if options.multiuser:
		subdir = os.path.join(options.csv_dir, key)
		os.makedirs(subdir, exist_ok=True)
		file_path = os.path.join(subdir, filename)
	else:
		file_path = os.path.join(options.csv_dir, filename)
	return file_path
	

def do_export_work_dir(connection, session, work_dir, options):
	'''Export a working directory of image redictions to a single file'''
	fieldnames = ["session","observer","organization","location","type"]
	fieldnames.extend(EXPORT_HEADERS)
	if not session_processed(connection, session):
		log.info("No new CSV file generation")
		return

	for (night,) in night_iterable(connection, session):
		# Write a session CSV file
		session_csv_file = get_file_path(connection, night, work_dir, options)
		with myopen(session_csv_file, 'w') as csvfile:
			writer = csv.writer(csvfile, delimiter=';')
			writer.writerow(fieldnames)
			for row in export_session_iterable(connection, session, night):
				row = map(var2std, enumerate(row))
				writer.writerow(row)
	log.info("Saved data to session CSV file {0}".format(session_csv_file))
	
	

def do_export_all(connection,  options):
	'''Exports all the database to a single file'''
	fieldnames = ["session","observer","organization","location","type"]
	fieldnames.extend(EXPORT_HEADERS)
	with myopen(options.csv_file, 'w') as csvfile:
		writer = csv.writer(csvfile, delimiter=';')
		writer.writerow(fieldnames)
		for row in export_all_iterable(connection):
			row = map(var2std, enumerate(row))
			writer.writerow(row)
	log.info("Saved data to global CSV file {0}".format(options.csv_file))

# ==================================
# Image List subcommands and options
# ==================================


EXIF_HEADERS = [
	'Name',
	'Session',
	'Timestamp',
	'Model',
	'Exposure',
	'ISO',
	'Focal',
	'f/'
]

GLOBAL_HEADERS = [
	'Name',
	'Type',
	'Session',
	'Observer',
	'Organiztaion',
	'Location',
	'ROI',
]

STATE_HEADERS = [
	"Name",
	"Session",
	"Type", 
	"State",
]

DATA_HEADERS = [
	"Name", "ROI", "Bias",
	"\u03BC R1", "\u03C3^2 R1", 
	"\u03BC G2", "\u03C3^2 G2", 
	"\u03BC G3", "\u03C3^2 G3",
	"\u03BC B4", "\u03C3^2 B4",
]

RAW_DATA_HEADERS = [
	"Name", "ROI" , "Bias",
	"Raw \u03BC R1", "Raw \u03C3^2 R1", 
	"Raw \u03BC G2", "Raw \u03C3^2 G2", 
	"Raw \u03BC G3", "Raw \u03C3^2 G3",
	"Raw \u03BC B4", "Raw \u03C3^2 B4",
]

DARK_DATA_HEADERS = [
	"Name", "ROI" , "Bias",
	"Dark \u03BC R1", "Dark \u03C3^2 R1", 
	"Dark \u03BC G2", "Dark \u03C3^2 G2", 
	"Dark \u03BC G3", "Dark \u03C3^2 G3",
	"Dark \u03BC B4", "Dark \u03C3^2 B4",
]

def view_session_count(cursor, session):
	row = {'session': session}
	cursor.execute(
		'''
		SELECT COUNT(*)
		FROM image_t
		WHERE session = :session
		''', row)
	return cursor.fetchone()[0]


def view_all_count(cursor):
	cursor.execute(
		'''
		SELECT COUNT(*)
		FROM image_t
		''')
	return cursor.fetchone()[0]

# --------------
# Image metadata
# --------------

def view_meta_exif_all_iterable(connection, session):
	cursor = connection.cursor()
	count = view_all_count(cursor)
	cursor.execute(
		'''
		SELECT name, session, tstamp, model, exptime, iso, focal_length, f_number
		FROM image_t
		ORDER BY session DESC, name ASC
		''')
	return cursor, count


def view_meta_exif_session_iterable(connection, session):
	'''session may be None for NULL'''
	row = {'session': session}
	cursor = connection.cursor()
	count = view_session_count(cursor, session)
	cursor.execute(
		'''
		SELECT name, session, tstamp, model, exptime, iso, focal_length, f_number
		FROM image_t
		WHERE session = :session
		ORDER BY name DESC
		''', row)
	return cursor, count

# ------------
# Image General
# -------------

def view_meta_global_all_iterable(connection, session):
	cursor = connection.cursor()
	count = view_all_count(cursor)
	cursor.execute(
		'''
		SELECT name, type, session, observer, organization, email, location, roi
		FROM image_t
		ORDER BY session DESC
		''')
	return cursor, count


def view_meta_global_session_iterable(connection, session):
	'''session may be None for NULL'''
	row = {'session': session}
	cursor = connection.cursor()
	count = view_session_count(cursor, session)
	cursor.execute(
		'''
		SELECT name, type, session, observer, organization, email, location, roi
		FROM image_t
		WHERE session = :session
		ORDER BY name ASC
		''', row)
	return cursor, count

# -----------
# Image State
# -----------

def view_state_session_iterable(connection, session):
	row = {'session': session}
	cursor = connection.cursor()
	count = view_session_count(cursor, session)
	cursor.execute(
		'''
		SELECT name, session, type, s.label
		FROM image_t
		JOIN state_t AS s USING(state)
		WHERE session = :session
		ORDER BY session DESC, name ASC
		''', row)
	return cursor, count


def view_state_all_iterable(connection, session):
	row = {'session': session}
	cursor = connection.cursor()
	count = view_all_count(cursor)
	cursor.execute(
		'''
		SELECT name, session, type, s.label
		FROM image_t
		JOIN state_t AS s USING(state)
		ORDER BY session DESC, name ASC
		''', row)
	return cursor, count

# -----------
# Image Data
# -----------

def view_data_session_iterable(connection, session):
	row = {'session': session}
	cursor = connection.cursor()
	count = view_session_count(cursor, session)
	cursor.execute(
		'''
		SELECT 
			name, roi, bias,
			aver_signal_R1, vari_signal_R1,
			aver_signal_G2, vari_signal_G2,
			aver_signal_G3, vari_signal_G3,
			aver_signal_B4, vari_signal_B4
		FROM image_v
		WHERE session = :session
		ORDER BY name ASC
		''', row)
	return cursor, count


def view_data_all_iterable(connection, session):
	cursor = connection.cursor()
	count = view_all_count(cursor)
	cursor.execute(
		'''
		SELECT 
			name, roi, bias,
			aver_signal_R1, vari_signal_R1,
			aver_signal_G2, vari_signal_G2,
			aver_signal_G3, vari_signal_G3,
			aver_signal_B4, vari_signal_B4
		FROM image_v
		ORDER BY session DESC, name ASC
		''', row)
	return cursor, count

# -------------
# Raw Image Data
# --------------

def view_raw_data_session_iterable(connection, session):
	row = {'session': session, 'light': LIGHT_FRAME, 'unknown': UNKNOWN}
	cursor = connection.cursor()
	count = view_session_count(cursor, session)
	cursor.execute(
		'''
		SELECT 
			name, roi, bias,
			aver_signal_R1, vari_signal_R1,
			aver_signal_G2, vari_signal_G2,
			aver_signal_G3, vari_signal_G3,
			aver_signal_B4, vari_signal_B4
		FROM image_t
		WHERE session = :session
		AND ((type = :light) OR (type = :unknown))
		ORDER BY name ASC
		''', row)
	return cursor, count


def view_raw_data_all_iterable(connection, session):
	row = {'light': LIGHT_FRAME, 'unknown': UNKNOWN}
	cursor = connection.cursor()
	count = view_all_count(cursor)
	cursor.execute(
		'''
		SELECT 
			name, roi, bias,
			aver_signal_R1, vari_signal_R1,
			aver_signal_G2, vari_signal_G2,
			aver_signal_G3, vari_signal_G3,
			aver_signal_B4, vari_signal_B4
		FROM image_t
		WHERE type = :type
		AND ((type = :light) OR (type = :unknown))
		ORDER BY session DESC, name ASC
		''', row)
	return cursor, count

# --------------
# Dark Image Data
# ---------------

def view_dark_data_session_iterable(connection, session):
	row = {'session': session}
	cursor = connection.cursor()
	count = view_session_count(cursor, session)
	cursor.execute(
		'''
		SELECT 
			name, roi, bias
			aver_dark_R1, vari_dark_R1,
			aver_dark_G2, vari_dark_G2,
			aver_dark_G3, vari_dark_G3,
			aver_dark_B4, vari_dark_B4
		FROM image_t
		WHERE session = :session
		ORDER BY name ASC
		''', row)
	return cursor, count


def view_dark_data_all_iterable(connection, session):
	cursor = connection.cursor()
	count = view_all_count(cursor)
	cursor.execute(
		'''
		SELECT 
			name, roi, bias
			aver_dark_R1, vari_dark_R1,
			aver_dark_G2, vari_dark_G2,
			aver_dark_G3, vari_dark_G3,
			aver_dark_B4, vari_dark_B4
		FROM image_t
		ORDER BY session DESC, name ASC
		''', row)
	return cursor, count

# ----------------
# View Master Dark
# -----------------

def view_master_dark_all_iterable(connection, session):
	row = {'tolerance': 0.2}
	cursor = connection.cursor()
	cursor.execute("SELECT COUNT(*) FROM master_dark_t")
	count = cursor.fetchone()[0]
	cursor.execute(
		'''
		SELECT 
			session, N, roi,
			(max_exptime - min_exptime) <= :tolerance as good_flag,            
			aver_R1, vari_R1,             
			aver_G2, vari_G2,         
			aver_G3, vari_G3,             
			aver_B4, vari_B4             
		FROM master_dark_t
		ORDER BY session DESC
		''', row)
	return cursor, count


def view_master_dark_session_iterable(connection, session):
	row = {'tolerance': 0.2, 'session': session}
	cursor = connection.cursor()
	cursor.execute("SELECT COUNT(*) FROM master_dark_t WHERE session = :session", row)
	count = cursor.fetchone()[0]
	cursor.execute(
		'''
		SELECT 
			session, N, roi,         
			(max_exptime - min_exptime) <= :tolerance as good_flag,
			aver_R1, vari_R1,             
			aver_G2, vari_G2,         
			aver_G3, vari_G3,             
			aver_B4, vari_B4
		FROM master_dark_t
		WHERE session = :session
		''', row)
	return cursor, count


MASTER_DARK_HEADERS = [
	"Session", 
	"# Darks",
	"ROI",
	"Good?",
	"\u03BC R1", "\u03C3^2 R1", 
	"\u03BC G2", "\u03C3^2 G2", 
	"\u03BC G3", "\u03C3^2 G3",
	"\u03BC B4", "\u03C3^2 B4",
]


# ---------
# View Dark
# ----------

def view_dark_session_iterable(connection, session):
	row = {'session': session, 'type': DARK_FRAME}
	cursor = connection.cursor()
	count = view_session_count(cursor, session)
	cursor.execute(
		'''
		SELECT 
			name, roi, bias,
			aver_signal_R1, vari_signal_R1,
			aver_signal_G2, vari_signal_G2,
			aver_signal_G3, vari_signal_G3,
			aver_signal_B4, vari_signal_B4
		FROM image_t
		WHERE session = :session
		AND type = :type
		ORDER BY name ASC
		''', row)
	return cursor, count


def view_dark_all_iterable(connection, session):
	row = {'session': session, 'type': DARK_FRAME}
	cursor = connection.cursor()
	count = view_all_count(cursor)
	cursor.execute(
		'''
		SELECT 
			name, roi, bias,
			aver_signal_R1, vari_signal_R1,
			aver_signal_G2, vari_signal_G2,
			aver_signal_G3, vari_signal_G3,
			aver_signal_B4, vari_signal_B4
		FROM image_t
		WHERE type = :type
		ORDER BY session DESC, name ASC
		''', row)
	return cursor, count


def do_image_view(connection, session, iterable, headers, options):
	cursor, count = iterable(connection, session)
	paging(cursor, headers, maxsize=count, page_size=options.page_size)

# =====================
# Command esntry points
# =====================

# These display various data

def image_list(connection, options):
	session = latest_session(connection)
	if options.exif:
		headers = EXIF_HEADERS
		iterable = view_meta_exif_all_iterable if options.all else view_meta_exif_session_iterable
	elif options.generic:
		headers = GLOBAL_HEADERS
		iterable = view_meta_global_all_iterable if options.all else view_meta_global_session_iterable
	elif options.state:
		headers = STATE_HEADERS
		iterable = view_state_all_iterable if options.all else view_state_session_iterable
	elif options.data:
		headers = DATA_HEADERS
		iterable = view_data_all_iterable if options.all else view_data_session_iterable
	elif options.raw_data:
		headers = RAW_DATA_HEADERS
		iterable = view_raw_data_all_iterable if options.all else view_raw_data_session_iterable
	elif options.dark_data:
		headers = DARK_DATA_HEADERS
		iterable = view_dark_data_all_iterable if options.all else view_dark_data_session_iterable
	elif options.dark:
		headers = RAW_DATA_HEADERS
		iterable = view_dark_all_iterable if options.all else view_dark_session_iterable
	elif options.master:
		headers = MASTER_DARK_HEADERS
		iterable = view_master_dark_all_iterable if options.all else view_master_dark_session_iterable
	else:
		return
	do_image_view(connection, session, iterable, headers, options)


def image_export(connection, options):
	do_export_all(connection, options)
	

def do_image_reduce(connection, options):
	log.info("#"*48)
	tmp  = os.path.basename(options.work_dir)
	if tmp == '':
		options.work_dir = options.work_dir[:-1]
	log.info("Working Directory: %s", options.work_dir)
	file_options = load_config_file(options.config)
	options      = merge_options(options, file_options)
	session = work_dir_to_session(connection, options.work_dir, options.filter)
	if session is None:
		session = int(datetime.datetime.utcnow().strftime("%Y%m%d%H%M%S"))
		log.info("Start with new reduction session %d", session)
	else:
		log.info("Start with existing reduction session %d", session)

	# Step 1: registering
	register_deleted = do_register(connection, options.work_dir, options.filter, session)
	
	if options.reset:
		image_session_state_reset(connection, session)
	
	# Step 2
	try:
		stats_computed = do_stats(connection, session, options.work_dir, options)
	except MetadataError as e:
		log.error(e)
		work_dir_cleanup(connection)
		raise
	except ConfigError as e:
		log.error(e)
		work_dir_cleanup(connection)
		raise

	# Step 3
	metadata_updated = do_metadata(connection, session, options)

	# Step 4
	do_classify(connection, session, options.work_dir, options)

	# Step 5
	do_apply_dark(connection, session, options)

	# Step 6
	if register_deleted or stats_computed or metadata_updated or options.force_csv:
		try:
			do_export_work_dir(connection, session, options.work_dir, options)
		except IOError as e:
			log.error(e)
			work_dir_cleanup(connection)
			raise
	else:
		log.info("NO CSV file generation is needed for session %d", session)

	# Cleanup session stuff
	work_dir_cleanup(connection)


def do_image_multidir_reduce(connection, options):
	with os.scandir(options.work_dir) as it:
		dirs  = [ entry.path for entry in it if entry.is_dir()  ]
		files = [ entry.path for entry in it if entry.is_file() ]
	if dirs:
		if files:
			log.warning("Ignoring files in %s", options.work_dir)
		for item in sorted(dirs, reverse=True):
			options.work_dir = item
			try:
				do_image_reduce(connection, options)
			except ConfigError as e:
				pass
			time.sleep(1.5)
	else:
		do_image_reduce(connection, options)


def image_reduce(connection, options):
	
	def by_name(item):
		return item[0]

	if not options.multiuser:
		do_image_multidir_reduce(connection, options)
	else:
		# os.scandir() only available from Python 3.6   
		with os.scandir(options.work_dir) as it:
			dirs = [ (entry.name, entry.path) for entry in it if entry.is_dir() ]
		if dirs:
			for key, path in sorted(dirs, key=by_name):
				options.config   = os.path.join(AZOTEA_CFG_DIR, key + '.ini')
				options.work_dir = path
				try:
					do_image_multidir_reduce(connection, options)
				except IOError as e:
					log.warning("No %s.ini file, skipping observer", key)
		else:
			raise NoUserInfoError(options.work_dir)



# ----------------------------------------------------------------------
# Copyright (c) 2014 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------


#--------------------
# System wide imports
# -------------------

import os.path
import errno
import logging

import configparser as ConfigParser


#--------------
# local imports
# -------------


# ----------------
# Module constants
# ----------------

# -----------------------
# Module global variables
# -----------------------

log = logging.getLogger("azotenodo")

# ------------------------
# Module Utility Functions
# ------------------------

def load_config_file(filepath):
    '''
    Load options from configuration file whose path is given
    Returns a dictionary
    '''
    if filepath is None or not (os.path.exists(filepath)):
        raise IOError(errno.ENOENT,"No such file or directory", filepath)

    parser  = ConfigParser.RawConfigParser()
    # str is for case sensitive options
    parser.optionxform = str
    parser.read(filepath)
    log.info("Opening configuration file %s", filepath)

    options = {}
    options['api_key_sandbox']    = parser.get("zenodo","api_key_sandbox")
    options['api_key_production'] = parser.get("zenodo","api_key_production")

    return options




# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------
# Copyright (c) 2020
#
# See the LICENSE file for details
# see the AUTHORS file for authors
# ----------------------------------------------------------------------

#--------------------
# System wide imports
# -------------------

import logging
import requests
import argparse
import os.path
import pprint
import json
import sqlite3

#--------------
# local imports
# -------------

from .packer import make_new_release
from . import SANDBOX_DOI_PREFIX, SANDBOX_URL_PREFIX, PRODUCTION_URL_PREFIX, PRODUCTION_DOI_PREFIX, DEF_DBASE
# -----------------------
# Module global variables
# -----------------------

log     = logging.getLogger("azotenodo")

# -----------------------
# Module global functions
# -----------------------

def setup_context(options, file_options):
    context = argparse.Namespace()

    if options.test:
        context.url_prefix = SANDBOX_URL_PREFIX
        context.doi_prefix = SANDBOX_DOI_PREFIX
        context.access_token = file_options.api_key_sandbox
    else:
        context.url_prefix = PRODUCTION_URL_PREFIX
        context.doi_prefix = PRODUCTION_DOI_PREFIX
        context.access_token = file_options.api_key_production
    return context


def open_database(dbase_path):
    if not os.path.exists(dbase_path):
        with open(dbase_path, 'w') as f:
            pass
        log.info("Created database file {0}".format(dbase_path))
    return sqlite3.connect(dbase_path)


def select_contributors(connection):
    cursor = connection.cursor()
    cursor.execute(
        '''
        SELECT DISTINCT obs_surname, obs_family_name, organization
        FROM image_t
        WHERE obs_surname IS NOT NULL
        AND obs_family_name IS NOT NULL
        ORDER BY obs_surname ASC
        ''')
    return cursor


def get_contributors(connection):
    contributors = []
    for surname, family_name, organization in select_contributors(connection):
        record = {
            'name': "{0}, {1}".format(surname, family_name),
            'type': 'DataCollector',
            }
        if organization is not None:
            record['affiliation'] = organization
        contributors.append(record)
    return contributors


# ------------
# Real actions
# ------------

def do_zenodo_licenses(context):
    headers = {"Content-Type": "application/json"}
    params  = {'access_token': context.access_token,}
    url = "{0}/licenses/".format(context.url_prefix)
    log.debug("Licenses List Request to {0} ".format(url))
    r = requests.get(url, params=params, headers=headers)
    log.info("Licenses List Status Code {0} ".format(r.status_code))
    response = r.json()
    if context.verbose:
        print("="*80)
        context.pprinter.pprint(response)
        print("="*80)
    if 400 <= r.status_code <= 599:
        raise Exception(response)
    return response


def do_zenodo_list(context, title, published):
    headers = {"Content-Type": "application/json"}
    status  = 'published' if published else 'draft' 
    query = 'type:dataset AND title:{0}'.format(title)
    query = 'title:{0}'.format(title)
    
    params  = {'access_token': context.access_token, 'status':status, 'sort': 'mostrecent', 'q': query}
    url = "{0}/deposit/depositions".format(context.url_prefix)
    
    log.debug("Deposition List Request to {0} ".format(url))
    r = requests.get(url, params=params, headers=headers)
    log.info("Deposition List Status Code {0} ".format(r.status_code))
    response = r.json()
    if context.verbose:
        print("=============== BEGIN DEPOSIT LISTING RESPONSE ===============")
        context.pprinter.pprint(response)
        print("=============== END DEPOSIT LISTING RESPONSE ===============")
    if 400 <= r.status_code <= 599:
        raise Exception(response)
    return response

def do_zenodo_search(context, title, published):
    headers = {"Content-Type": "application/json"}
    status  = 'published'
    params  = {'access_token': context.access_token, 'status':status, 'sort': 'mostrecent',}
    url = "{0}/deposit/depositions".format(context.url_prefix)
    
    log.info("Searching dataset with title {0}".format(title))
    log.debug("Deposition List Request to {0} ".format(url))
    r = requests.get(url, params=params, headers=headers)
    response = r.json()
    response = list(filter(lambda item: item['title'] == title, response))
    log.info("Deposition search OK, HTTP status code {0} ".format(r.status_code))
    if context.verbose:
        print("=============== BEGIN DEPOSITION SEARCH BY TITLE RESPONSE ===============")
        context.pprinter.pprint(response)
        print("=============== END DEPOSITION SEARCH BY TITLE RESPONSE ===============")
    if 400 <= r.status_code <= 599:
        raise Exception(response)
    return response[0]


def do_zenodo_delete(context, identifier):
    headers = {"Content-Type": "application/json"}
    params  = {'access_token': context.access_token}
    url = "{0}/deposit/depositions/{1}".format(context.url_prefix, identifier)
    log.debug("Deposition Delete  Request to {0} ".format(url))
    r = requests.delete(url, params=params, headers=headers)
    log.info("Deposition Delete Status Code {0} ".format(r.status_code))
    response = r.json()
    if context.verbose:
        print("=============== BEGIN DEPOSIT DELETION RESPONSE ===============")
        context.pprinter.pprint(response)
        print("=============== END DEPOSIT DELETETION RESPONSE ===============")
    if 400 <= r.status_code <= 599:
        raise Exception(response)
    return response


def do_zenodo_deposit(context):
    headers = {"Content-Type": "application/json"}
    params  = {'access_token': context.access_token}
    log.info("Creating new Deposition for title {0}, version {1} ".format(context.title, context.version))
    url = "{0}/deposit/depositions".format(context.url_prefix)
    log.debug("Deposition Request to {0} ".format(url))
    r = requests.post(url, params=params, headers=headers, json={})
    response = r.json()
    if context.verbose:
        print("=============== BEGIN DEPOSIT CREATION RESPONSE ===============")
        context.pprinter.pprint(response)
        print("=============== END DEPOSIT CREATION RESPONSE ===============")
    if 400 <= r.status_code <= 599:
        raise Exception(response)
    log.info("Deposition created with id {0}, HTTP status code {1}".format(response['id'], r.status_code))
    return response


def do_zenodo_metadata(context, identifier):
    log.info("Deposit Metadata for id {0} to Zenodo".format(identifier))

    connection   = open_database(DEF_DBASE)
    contributors = get_contributors(connection)

    headers = {"Content-Type": "application/json"}
    params  = {'access_token': context.access_token}
    metadata = {
        'title' : context.title,
        'upload_type': 'dataset',
        'version' : context.version,
        'communities': [ {'identifier': context.community} ],
        'creators' : [
            {'name': 'Zamorano, Jaime', 'affiliation': 'UCM', 'orcid': 'https://orcid.org/0000-0002-8993-5894'},
            {'name': 'González, Rafael','affiliation': 'UCM', 'orcid': 'https://orcid.org/0000-0002-3725-0586'}
        ],
        'contributors': contributors,
        'description': 'Latest monthly AZOTEA reduced CSV files',
        'access_right': 'open',
    }
    url = "{0}/deposit/depositions/{1}".format(context.url_prefix, identifier)
    log.debug("Deposition Metadata Request to {0} ".format(url))
    r = requests.put(url, params=params, headers=headers, json={'metadata':metadata})
    response = r.json()
    if context.verbose:
        print("=============== BEGIN METADATA RESPONSE ===============")
        context.pprinter.pprint(response)
        print("=============== END METADATA RESPONSE ===============")
    if 400 <= r.status_code <= 599:
        raise Exception(response)
    log.info("Metadata updated for id {0}, HTTP status code {1}".format(identifier, r.status_code))
    return response


def do_zenodo_upload(context, zip_file, bucket_url):
    headers = {"Content-Type": "application/json"}
    params  = {'access_token': context.access_token}
    filename   = os.path.basename(zip_file)
    url = "{0}/{1}".format(bucket_url, filename)
    with open(zip_file, "rb") as fp:
        log.debug("Deposition File Upload Request to {0} ".format(url))
        r = requests.put(url, data=fp, params=params)
    response = r.json()
    if context.verbose:
        print("=============== BEGIN FILE UPLOAD RESPONSE ===============")
        context.pprinter.pprint(response)
        print("=============== END FILE UPLOAD RESPONSE ===============")
    if 400 <= r.status_code <= 599:
        raise Exception(response)
    log.info("Deposition File Upload succesful for {0}, HTTP status code {1} ".format(zip_file, r.status_code))
    return response


def do_zenodo_publish(context, identifier):
    headers = {"Content-Type": "application/json"}
    params  = {'access_token': context.access_token}
    url = "{0}/deposit/depositions/{1}/actions/publish".format(context.url_prefix, identifier)
    log.info("Publish new dataset for {0}".format(identifier))
    log.debug("Deposition Publish Request to {0} ".format(url))
    r = requests.post(url, params=params, headers=headers, json={})
    response = r.json()
    if context.verbose:
        print("=============== BEGIN PUBLISH RESPONSE ===============")
        context.pprinter.pprint(response)
        print("=============== END PUBLISH RESPONSE ===============")
    if 400 <= r.status_code <= 599:
        raise Exception(response)
    log.info("Publication succesful doi = {0}, HTTP status code {1} ".format(response['doi'], r.status_code))
    return context


def do_zenodo_newversion(context, latest_id):
    headers = {"Content-Type": "application/json"}
    params  = {'access_token': context.access_token,}
    url = "{0}/deposit/depositions/{1}/actions/newversion".format(context.url_prefix, latest_id)
    log.info("Creating New Version deposition from {0}".format(latest_id))
    log.debug("Deposition New Version of {1} Request to {0} ".format(url, latest_id))
    r = requests.post(url, params=params, headers=headers, json={})
    response = r.json()
    if context.verbose:
        print("=============== BEGIN DEPOSITION NEW VERSION RESPONSE ===============")
        context.pprinter.pprint(response)
        print("=============== END DEPOSITION NEW VERSION RESPONSE ===============")
    if 400 <= r.status_code <= 599:
        raise Exception(response)
    new_id   = os.path.basename(response['links']['latest_draft'])
    log.info("Deposition New Version succesful, new id = {0}, HTTP status code {1} ".format(new_id, r.status_code))
    return response

# ========
# COMMANDS
# ========

def zenodo_licenses(options, file_options):
    context = setup_context(options, file_options)
    context.verbose  = options.verbose
    context.pprinter = pprint.PrettyPrinter(indent=2)
    do_zenodo_licenses(context)


def zenodo_list(options, file_options):
    context = setup_context(options, file_options)
    context.verbose   = options.verbose
    context.pprinter  = pprint.PrettyPrinter(indent=2)
    do_zenodo_list2(context, options.title, options.published)


def zenodo_delete(options, file_options):
    context = setup_context(options, file_options)
    context.verbose  = options.verbose
    context.pprinter = pprint.PrettyPrinter(indent=2)
    identifier       = options.id
    do_zenodo_delete(context, identifier)


def zenodo_pipeline(options, file_options):
    first_time, changed, version = make_new_release(options)
    if not changed:
        log.info("No need to upload new version to Zendodo")
        return

    context = setup_context(options, file_options)
    context.verbose   = options.verbose
    context.pprinter  = pprint.PrettyPrinter(indent=2)
    context.title     = options.title
    context.community = options.community
    context.version   = version if options.version is None else options.version
    zip_file          = options.zip_file
   
    if first_time:
        response = do_zenodo_deposit(context)
        new_id = response['id']
        response = do_zenodo_metadata(context, new_id)
        bucket_url = response["links"]["bucket"]
        response = do_zenodo_upload(context, zip_file, bucket_url)
        response = do_zenodo_publish(context, new_id)
    else:
        response = do_zenodo_search(context, context.title, True)
        latest_id = response['id']
        response = do_zenodo_newversion(context, latest_id)
        new_id   = os.path.basename(response['links']['latest_draft'])
        response = do_zenodo_metadata(context, new_id)
        bucket_url = response["links"]["bucket"]
        response = do_zenodo_upload(context, zip_file, bucket_url)
        response = do_zenodo_publish(context, new_id)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------
# Copyright (c) 2020
#
# See the LICENSE file for details
# see the AUTHORS file for authors
# ----------------------------------------------------------------------

#--------------------
# System wide imports
# -------------------

import os.path
import logging
import shutil
import datetime
import glob

# ---------------------
# Third party libraries
# ---------------------

#--------------
# local imports
# -------------

from . import  *

# ----------------
# Module constants
# ----------------

# -----------------------
# Module global variables
# -----------------------

log = logging.getLogger("azotea")

# -----------------------
# Module global functions
# -----------------------

def latest_session(connection):
    '''Get Last recorded session'''
    cursor = connection.cursor()
    cursor.execute('''
        SELECT MAX(session)
        FROM image_t 
        ''')
    return cursor.fetchone()[0]


def dbase_do_backup(comment):
    tstamp = datetime.datetime.utcnow().strftime(".%Y%m%d%H%M%S")
    filename = os.path.basename(DEF_DBASE) + tstamp
    dest_file = os.path.join(AZOTEA_BAK_DIR, filename)
    shutil.copy2(DEF_DBASE, dest_file)
    log.info("database backup to {0}".format(dest_file))


def dbase_delete_selected_images(connection, session):
    row = {'session': session}
    cursor = connection.cursor()
    cursor.execute('''
        DELETE FROM image_t
        WHERE session == :session 
        ''', row)

def dbase_delete_selected_master_dark(connection, session):
    row = {'session': session}
    cursor = connection.cursor()
    cursor.execute('''
        DELETE FROM master_dark_t
        WHERE session == :session 
        ''', row)

# =====================
# Command esntry points
# =====================


def database_clear(connection, options):
    cursor = connection.cursor()
    if options.all:
        cursor.execute("DELETE FROM image_t")
        cursor.execute("DELETE FROM master_dark_t")
        log.info("Cleared all data from database {0}".format(os.path.basename(DEF_DBASE)))
    else:
        session = latest_session(connection)
        dbase_delete_selected_master_dark(connection, session)
        dbase_delete_selected_images(connection, session)
        log.info("Cleared data from session {1} in database {0}, ".format(os.path.basename(DEF_DBASE), session))
    connection.commit()



def database_purge(connection, options):
    cursor = connection.cursor()
    with open(SQL_PURGE) as f: 
        lines = f.readlines() 
    script = ''.join(lines)
    connection.executescript(script)
    connection.commit()    
    log.info("Erased schema in database {0}".format(os.path.basename(DEF_DBASE)))



def database_backup(connection, options):
    connection.close()
    dbase_do_backup(options.comment)


# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------
# Copyright (c) 2020
#
# See the LICENSE file for details
# see the AUTHORS file for authors
# ----------------------------------------------------------------------

#--------------------
# System wide imports
# -------------------

import sys
import os.path
import logging
import shutil
import difflib

# ---------------------
# Third party libraries
# ---------------------


#--------------
# local imports
# -------------

from . import DEF_CAMERA_TPL, DEF_CONFIG_TPL, AZOTEA_CFG_DIR

# ----------------
# Module constants
# ----------------

# -----------------------
# Module global variables
# -----------------------

log = logging.getLogger("azotea")

# -----------------------
# Module global functions
# -----------------------

def config_list(filename):
	print('\n')
	with open(filename, 'r') as f:
		shutil.copyfileobj(f, sys.stdout)


def config_create(filename):
	dest = os.path.join(AZOTEA_CFG_DIR, os.path.basename(filename))
	shutil.copy2(filename, dest)
	log.info("Created {0} file".format(dest))


def config_diff(template_file, input_file):
	with open(template_file,'r') as tpl_file:
		tpl_contents = tpl_file.readlines()
	with open(input_file,'r') as inp_file:
		inp_contents = inp_file.readlines()
	tpl_contents = [line.rstrip() for line in tpl_contents]
	inp_contents = [line.rstrip() for line in inp_contents]
	for line in difflib.context_diff(tpl_contents, inp_contents, 
			fromfile=AZOTEA_CFG_DIR, tofile=input_file):
		print(line)


# =====================
# Command esntry points
# =====================


def config_global(connection, options):
	if options.list:
		config_list(DEF_CONFIG_TPL)
	elif options.create:
		config_create(DEF_CONFIG_TPL)
	elif options.diff:
		config_diff(DEF_CONFIG_TPL, options.input_file)


def config_camera(connection, options):
	if options.list:
		config_list(DEF_CAMERA_TPL)
	elif options.create:
		config_create(DEF_CAMERA_TPL)



import os
import os.path

from setuptools import setup, Extension
import versioneer

# Default description in markdown
LONG_DESCRIPTION = open('README.md').read()


PKG_NAME     = 'azotea'
AUTHOR       = 'Rafael González'
AUTHOR_EMAIL = 'astrorafael@gmail.com'
DESCRIPTION  = 'command line tool to reduce DSLR RAW images',
LICENSE      = 'MIT'
KEYWORDS     = 'Astronomy Python LightPollution'
URL          = 'https://guaix.ucm.es/AZOTEA'
PACKAGES     = ["azotea","azotenodo"]
DEPENDENCIES = [
                  'tabulate',
                  'numpy',
                  'matplotlib',
                  'rawpy',
                  'exifread',
                  'opencv-python',
                  'jdcal',
                  'requests'
]

CLASSIFIERS  = [
    'Environment :: Console',
    'Intended Audience :: Science/Research',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Operating System :: POSIX :: Linux',
    'Programming Language :: Python :: 3.6',
    'Topic :: Scientific/Engineering :: Astronomy',
    'Topic :: Scientific/Engineering :: Atmospheric Science',
    'Development Status :: 4 - Beta',
]

# Additional data inside the package
PACKAGE_DATA = {
    'azotea': [
                'data/camera.ini',
                'data/azotea.ini',
                'data/sql/*.sql',
                'data/sql/data/*.sql',
              ],
    'azotenodo': [
                'data/azotenodo.ini',
              ],
}

SCRIPTS = [
    'files/azotea',
    'files/migrate_hash',
]



setup(
    name             = PKG_NAME,
    version          = versioneer.get_version(),
    cmdclass         = versioneer.get_cmdclass(),
    author           = AUTHOR,
    author_email     = AUTHOR_EMAIL,
    description      = DESCRIPTION,
    long_description_content_type = "text/markdown",
    long_description = LONG_DESCRIPTION,
    license          = LICENSE,
    keywords         = KEYWORDS,
    url              = URL,
    classifiers      = CLASSIFIERS,
    packages         = PACKAGES,
    install_requires = DEPENDENCIES,
    package_data     = PACKAGE_DATA,
    scripts          = SCRIPTS
)
 



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------
# Copyright (c) 2020.
#
# See the LICENSE file for details
# see the AUTHORS file for authors
# ----------------------------------------------------------------------

#--------------------
# System wide imports
# -------------------

import os
import os.path

# Access  template withing the package
from pkg_resources import resource_filename

#--------------
# local imports
# -------------

from ._version import get_versions

# ----------------
# Module constants
# ----------------

# ----------------
# Module constants
# ----------------

PRODUCTION_DOI_PREFIX = "10.5281"
SANDBOX_DOI_PREFIX    = "10.5072"

PRODUCTION_URL_PREFIX = "https://zenodo.org/api"
SANDBOX_URL_PREFIX    = "https://sandbox.zenodo.org/api"

AZOTEA_PUBL_TITLE     = 'AZOTEA dataset'
AZOTEA_COMMUNITY      = 'azotea'


# Configuration file templates are built-in the package
DEF_CONFIG_TPL = resource_filename(__name__, os.path.join('data', 'azotenodo.ini'))

# Configuration file templates are built-in the package
AZOTEA_BASE_DIR = os.path.join(os.path.expanduser("~"), "azotea")
AZOTEA_CFG_DIR  = os.path.join(AZOTEA_BASE_DIR, "config")
AZOTEA_LOG_DIR  = os.path.join(AZOTEA_BASE_DIR, "log")
AZOTEA_CSV_DIR  = os.path.join(AZOTEA_BASE_DIR, "csv")
AZOTEA_ZIP_DIR  = os.path.join(AZOTEA_BASE_DIR, "zip")
AZOTEA_DB_DIR   = os.path.join(AZOTEA_BASE_DIR, "dbase")

DEF_DBASE      = os.path.join(AZOTEA_DB_DIR, "azotea.db")
DEF_CONFIG     = os.path.join(AZOTEA_CFG_DIR, os.path.basename(DEF_CONFIG_TPL))
DEF_ZIP_FILE   = os.path.join(AZOTEA_ZIP_DIR, "azotea.zip")

DEF_TSTAMP = "%Y-%m-%dT%H:%M:%S"
# -----------------------
# Module global variables
# -----------------------

# Git Version Management
__version__ = get_versions()['version']
del get_versions



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------
# Copyright (c) 2020
#
# See the LICENSE file for details
# see the AUTHORS file for authors
# ----------------------------------------------------------------------

#--------------------
# System wide imports
# -------------------

import sys
import os
import os.path
import glob
import logging
import shutil
import subprocess

# ---------------------
# Third party libraries
# ---------------------

import jdcal

#--------------
# local imports
# -------------

from .camera import CameraImage
from .utils  import LogCounter

# ----------------
# Module constants
# ----------------

N_FILES = 50

# -----------------------
# Module global variables
# -----------------------

log = logging.getLogger("azotea")

# -----------------------
# Module global functions
# -----------------------

def _copyfileobj_patched(fsrc, fdst, length=16*1024*1024):
    """Patches shutil method to improve big file copy speed on Linux"""
    while 1:
        buf = fsrc.read(length)
        if not buf:
            break
        fdst.write(buf)


shutil.copyfileobj = _copyfileobj_patched


def dir_name(jd2000, mjd):
	year, month, day, fraction = jdcal.jd2gcal(jd2000, mjd)
	return "{0:04d}-{1:02d}-{2:02d}".format(year, month, day)

def scan_images(options):
	output_dir_set = set()
	image_list = []
	counter = LogCounter(N_FILES)
	filepath_iterable = glob.iglob(os.path.join(options.input_dir, '*'))
	for input_file_path in filepath_iterable:
		image = CameraImage(input_file_path, options)
		image.loadEXIF()
		jd2000, mjd = image.getJulianDate()
		mjd -= 0.5	# Take it 12 hours before and make sure it is the same YYYY-MM-DD
		date_string = dir_name(jd2000, mjd)
		output_dir_path = os.path.join(options.output_dir, date_string)
		output_dir_set.add(output_dir_path)
		image_list.append((input_file_path, output_dir_path))
		counter.tick("Read %d images")
	counter.end("Read %d images")
	return output_dir_set, image_list


def create_dest_directories(output_dir_set):
	log.info("creating %d output directories", len(output_dir_set))
	for directory in output_dir_set:
		if not os.path.isdir(directory):
			os.makedirs(directory)


def copy_files(image_list):
	log.info("copying images to output directories")
	counter = LogCounter(N_FILES)
	for item in image_list:
		if sys.platform == 'win32':
			subprocess.call("xcopy",item[0], item[1])
		else:
			shutil.copy2(item[0], item[1])
		counter.tick("Copied %d images")
	counter.end("Copied %d images")


# =====================
# Command esntry points
# =====================



def reorganize_images(connection, options):
	connection.close()
	output_dir_set, image_list = scan_images(options)
	if not options.dry_run:
		create_dest_directories(output_dir_set)
		copy_files(image_list)
	else:
		log.info("Would create %d directories", len(output_dir_set))



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------
# Copyright (c) 2020
#
# See the LICENSE file for details
# see the AUTHORS file for authors
# ----------------------------------------------------------------------

#--------------------
# System wide imports
# -------------------

import logging
import os
import os.path
import zipfile
import hashlib
import datetime

#--------------
# local imports
# -------------


# -----------------------
# Module global variables
# -----------------------

SEMANTIC_VERSIONING_FMT = "%y.%m"

log     = logging.getLogger("azotenodo")

# -----------------------
# Module global functions
# -----------------------

def fingerprint(filepath):
    '''Compute a hash from the image'''
    file_hash = hashlib.md5()
    with open(filepath, 'rb') as f:
        block = f.read() 
        while len(block) > 0:
            file_hash.update(block)
            block = f.read()
    return file_hash.digest()


def get_paths(directory):
    '''Get all file paths in a list''' 
  
    file_paths = [] 
  
    # crawling through directory and subdirectories 
    for root, directories, files in os.walk(directory):
        log.info("Exploring = {0}".format(root))
        for filename in files: 
            filepath = os.path.join(root, filename) 
            file_paths.append(filepath) 
    return file_paths         


def pack(options):
    '''Pack all files in the ZIP file given by options'''
    paths = get_paths(options.csv_dir)
    log.info("Creating {0}".format(options.zip_file))
    with zipfile.ZipFile(options.zip_file, 'w') as myzip:
        for myfile in paths: 
            myzip.write(myfile) 



def make_new_release(options):
    hsh        = {}
    first_time = False
    version    = None
    changed    = False
    if os.path.exists(options.zip_file):
        hsh['prev'] = fingerprint(options.zip_file)
    else:
        hsh['prev'] = None
        first_time = True
    pack(options)
    hsh['now'] = fingerprint(options.zip_file)
    if hsh['now'] != hsh['prev']:
        log.info("new files were added to {0}".format(options.zip_file))
        version = datetime.datetime.utcnow().strftime(SEMANTIC_VERSIONING_FMT)
        changed = True
    return first_time, changed, version
